# Copyright (c) 2012-2014 Kapiche Limited
# Author: Kris Rogers <kris@kapiche.com>, Ryan Stuart <ryan@kapiche.com>
"""
An sqlite implementation of :class:`caterpillar.storage.Storage`.

The only class is :class:`.SqliteStorage` which uses sqlite in WAL mode to achieve reader/writer isolation.

All changes to an index are first staged to a temporary in-memory database, the main storage file is not
updated until commit is called. At this point all of the contents of the index are flushed to the file. By
leaving the final flush operation as a single SQL script we can drop the GIL and allow concurrent operation
in multiple threads.

Note that document deletes are 'soft' deletes. Wherever possible the document data is deleted, however in
the document_data and term_posting tables a hard delete requires a full table scan, so this is not ordinarily
performed.

"""
from __future__ import division
import logging
import math
import os

import apsw

from caterpillar.processing.index import NonIndexedFieldError
from caterpillar.storage import StorageWriter, StorageReader, Storage, StorageNotFoundError, \
    DuplicateStorageError, PluginNotFoundError

from ._sqlite_schema import disk_schema, cache_schema, prepare_flush, flush_cache


logger = logging.getLogger(__name__)


class SqliteWriter(StorageWriter):
    """
    This class uses SQLite to write data structures to disk.

    Reader / writer isolation here is provided by using `WAL mode <http://www.sqlite.org/wal.html>`_. There are no
    changes to the default checkpoint behaviour of SQLite, which at the time of writing defaults to 1000 pages.

    After initialisation all changes to the database are staged to a temporary in memory database. The changes are
    not flushed to persistent storage until the commit method of this storage object is called.

    """

    def __init__(self, path, create=False):
        """
        Initialise a new instance of this storage at ``path`` (str).

        If ``create`` (bool) is False and path doesn't exist then a :exc:`StorageNotFoundError` is raised. Otherwise,
        if ``create`` is True then we create the database if it doesn't already exist.

        If ``create`` is True and the DB already exist, then :exc:`DuplicateStorageError` is raised.

        """
        self._db_path = path
        self._db = os.path.join(path, 'storage.db')

        if not create and not os.path.exists(self._db):
            raise StorageNotFoundError('Can\'t find the resources required by SQLiteStorage. Is it corrupt?')
        elif create and os.path.exists(self._db):
            raise DuplicateStorageError('There already appears to be something stored at {}'.format(path))

        if create:
            connection = apsw.Connection(self._db, flags=apsw.SQLITE_OPEN_READWRITE | apsw.SQLITE_OPEN_CREATE)
            cursor = connection.cursor()

            # Setup schema and necessary pragmas (on disk). Note that the disk_schema script returns rows from the
            # pragma statements - the list call makes sure that all statements are run, even if we don't care about the
            # returned rows.
            list(cursor.execute(disk_schema))

            # The database is never written to directly, hence it is closed after initialisation.
            connection.close()

    def begin(self):
        """
        Begin a transaction.

        A temporary in-memory database is created to cache results, which is destroyed after the
        commit or rollback methods are called.

        """

        # If we're opening for writing, don't connect to the index directly.
        # Instead setup a temporary in memory database with the minimal schema.
        self._db_connection = apsw.Connection(':memory:')
        # We serialise writers during a write lock, and in normal cases the WAL mode avoids writers blocking
        # readers. Setting this is used to handle the one case in our normal operations that WAL mode requires
        # an exclusive lock for cleaning up the WAL file and associated shared-memory index.
        # See section 8 for the edge cases: https://www.sqlite.org/wal.html
        self._db_connection.setbusytimeout(1000)
        list(self._execute(cache_schema))
        self.doc_no = 0  # local only for this write transaction.
        self.frame_no = 0
        self.committed = False
        self._flushed = False

    def commit(self):
        """Commit a transaction.

        First the on disk database is attached and the current maximum document and frame ID's are returned.

        Then the content of the in memory cache is flushed to the database. The cache is then dropped. The begin
        method will need to be called before this method is usable for writing again.

        Returns a list of the added documents.

        """
        if not self._flushed:
            self._flush()
        self._execute('commit; detach database disk_index;')
        self.committed = True
        self.doc_no = 0
        self.frame_no = 0

        return self.__last_added_documents, self.__deleted_documents, self.__updated_plugins

    def rollback(self):
        """Rollback a transaction on an IndexWriter."""
        self._execute('rollback')
        self.doc_no = 0
        self.frame_no = 0

    def close(self):
        """
        Close this storage object and all its resources, rendering it UNUSABLE.

        This operates immediately: if the data has not been committed it will be destroyed.

        """
        # If we have committed, the resources have already been released.
        if not self.committed:
            self._db_connection.close()
            self._db_connection = None

    def _prepare_flush(self):
        """Prepare to flush the cached data to the index.

        This method returns the state of the index for synchronising document counts and a list of ID's
        for deleted documents.

        """
        return list(self._execute(prepare_flush, [self._db]))

    def _flush(self):
        """Actually perform the flush."""
        index_sync_data = self._prepare_flush()
        revision, max_document_id, deleted_count, max_frame_id = index_sync_data[0]
        self.__deleted_documents = index_sync_data[1:]
        self.__last_added_documents = list(range(max_document_id + 1, max_document_id + 1 + self.doc_no))
        self.__updated_plugins = list(
            self._execute(
                flush_cache,
                {
                    'max_doc': max_document_id + 1,
                    'max_frame': max_frame_id + 1,
                    'deleted': deleted_count + len(self.__deleted_documents),
                    'added': self.doc_no + max_document_id,
                    'added_frames': self.frame_no + max_frame_id
                })
        )
        self._flushed = True  # Only needed for _merge_term_variants currently.

    def add_structured_fields(self, field_names):
        """Register a structured field on the index. """
        for f in field_names:
            self._execute('insert into structured_field(name) values(?)', [f])

    def add_unstructured_fields(self, field_names):
        """Register an unstructured field on the index. """
        for f in field_names:
            self._execute('insert into unstructured_field(name) values(?)', [f])

    def delete_structured_fields(self, field_names):
        """Delete a structured field and the associated data from the index.

        Note this is a soft delete for the SqliteWriter class. Call
        :meth:SqliteWriter.materialize_deletes to remove all the data related
        to that field from the index. """
        raise NotImplementedError

    def delete_unstructured_fields(self, field_names):
        """Delete an unstructured field from the index.

        Note this is a soft delete for the SqliteWriter class. Call
        :meth:SqliteWriter.materialize_deletes to remove all the data related
        to that field from the index. """
        raise NotImplementedError

    def add_analyzed_document(self, document_format, document_data):
        """Add an analyzed document to the index.

        The added document will be assigned an integer document_id _after_ the commit method of this
        storage object runs to completion. These ID's are monotonically increasing and assigned in order
        of insertion.

        Arguments

            document_format: str
                A string representing the format of the passed data. Currently only 'v1'
                is supported.
            document_data:
                The data for the document, in the format expected for document_data.

        Valid Document Formats

            document_format == 'v1':
            An iterable of
                - a string representation of the whole document
                - a dictionary of field_name:field_value pairs for the document level structured data
                - a dictionary {
                    field_name: list of string representations of each frames
                }
                - a dictionary {
                    field_name: list of {term: [[word1 token_position], [word2 token_position]]} vectors for each frame
                }
            For the frame data (3rd and 4th elements), the frames should be in document sequence order
            and there should be a one-one correspondence between frame representations and term:frequency vectors.

        """
        if document_format == 'v1':
            try:
                # Create a savepoint so we don't have any problems with the field addition.
                self._execute('savepoint document')

                document, structured_data, frames, frame_terms = document_data

                # Stage the document.
                self._execute(
                    'insert into document(id, stored) values (?, ?)',
                    [self.doc_no, document]
                )

                # Stage the structured fields:
                insert_rows = ((self.doc_no, field, value) for field, value in structured_data.iteritems())
                self._executemany(
                    'insert into document_data(document_id, field_name, value) values (?, ?, ?)',
                    insert_rows
                )

                # Check frame data is consistent and pull out a frame count.
                number_frames = {field: len(values) for field, values in frames.iteritems()}
                number_frame_terms = {field: len(values) for field, values in frame_terms.iteritems()}

                if number_frames.keys() != number_frame_terms.keys():
                    raise ValueError('Inconsistent fields between frames and frame_terms')
                for field in number_frames:
                    if number_frames[field] != number_frame_terms[field]:
                        raise ValueError('Number of frames and frame_terms does not match for field {}'.format(field))

                total_frames = sum(number_frames.values())

                # Stage the frames:
                insert_frames = (
                    [self.doc_no, field, seq, frame]
                    for field, frame_list in sorted(frames.iteritems())
                    for seq, frame in enumerate(frame_list)
                )
                insert_frames_numbered = (
                    (frame_count + self.frame_no, a[0], a[1], a[2], a[3])
                    for frame_count, a in enumerate(insert_frames)
                )

                self._executemany(
                    'insert into frame(id, document_id, field_name, sequence, stored) values (?, ?, ?, ?, ?)',
                    insert_frames_numbered
                )

                # Term vectors for the frames, note that the dictionary is sorted for consistency with insert_frames
                frame_term_data = (
                    frame for field, frame_list in sorted(frame_terms.iteritems()) for frame in frame_list
                )
                insert_term_data = (
                    # The leading and trailing [] are stripped so positions can be concatenated as strings.
                    (frame_count + self.frame_no, term, len(positions), _bitwise_encode(positions))
                    for frame_count, frame_data in enumerate(frame_term_data)
                    for term, positions in frame_data.iteritems()
                )

                self._executemany(
                    'insert into stage_posting(frame_id, term, frequency, positions) values (?, ?, ?, ?)',
                    insert_term_data
                )

                self._execute('release document')  # rollup this savepoint into the transaction.
                self.frame_no += total_frames
                self.doc_no += 1

            except Exception as e:
                self._execute('rollback to savepoint document')
                raise e
        else:
            raise ValueError('Unknown document_format {}'.format(document_format))

    def append_frame_attributes(self, attribute_index):
        """Append the attributes for the given frames to the index. """
        row_generator = (
            (frame_id, attribute_type, attribute_value)
            for frame_id, values in attribute_index.iteritems()
            for attribute_type, attribute_value in values.iteritems()
        )
        self._executemany('insert into attribute_posting values (?, ?, ?)', row_generator)

    def delete_documents(self, document_ids):
        """Delete a document with the given id from the index. """
        document_id_gen = ((document_id,) for document_id in document_ids)
        self._executemany('insert into deleted_document(id) values(?)', document_id_gen)

    def set_plugin_state(self, plugin_type, plugin_settings, plugin_state):
        """ Set the plugin state in the index to the given state.

        Existing plugin state will be replaced.
        """

        # Insert into the plugin registry. If plugin_id already existed, reuse it.
        self._execute(
            "insert into plugin_registry(plugin_type, settings) values (?, ?); ",
            (plugin_type, plugin_settings)
        )

        insert_rows = ((plugin_type, plugin_settings, key, value) for key, value in plugin_state.iteritems())
        self._executemany("insert into plugin_data values (?, ?, ?, ?);", insert_rows)

    def delete_plugin_state(self, plugin_type, plugin_settings=None):
        """Delete a plugin instance, or all plugins of a certain type from the index. """
        self._execute('insert into delete_plugin values(?, ?)', (plugin_type, plugin_settings))

    def set_setting(self, name, value):
        """Set the setting ``name`` to ``value``"""
        self._execute('insert into setting values(?, ?)', [name, value])

    def _execute(self, query, data=None):
        cursor = self._db_connection.cursor()
        try:
            return cursor.execute(query, data)
        except apsw.SQLError as e:
            logger.exception(e)
            raise e

    def _executemany(self, query, data=None):
        cursor = self._db_connection.cursor()
        try:
            return cursor.executemany(query, data)
        except apsw.SQLError as e:
            logger.exception(e)
            raise e


class SqliteReader(StorageReader):
    """
    Reader class for data stored in SQLite by SQLiteWriter.

    A reader is transactionally isolated from writers by SQLite's Write Ahead Log. Calling the begin() method
    of this class begins a read transaction that does not end until commit is explicitly called.

    """
    def __init__(self, path):
        """Open or create a reader for the given storage location."""

        self._db_path = path
        self._db = os.path.join(path, 'storage.db')

        if not os.path.exists(self._db):
            raise StorageNotFoundError('Can\'t find the resources required by SQLiteStorage. Is it corrupt?')

        self._db_connection = apsw.Connection(self._db, flags=apsw.SQLITE_OPEN_READONLY)

        self._db_connection.setbusytimeout(1000)

    def begin(self):
        """Begin a read transaction."""
        self._execute('begin')
        # Temporary table for searches - only visible to this reader.
        self._execute("""
            create temporary table term_search_driver(
                term_id integer,  -- Drives the lookup in the index
                --These three columns drive the counts for deciding on matches
                all_id integer,
                n_id integer,
                exclude_count integer,
                -- Weighting for that particular term
                weight float default 1.0
            )
            """)

    def commit(self):
        """End the read transaction."""
        self._db_connection.cursor().execute('commit')
        return

    def close(self):
        """Close the reader, freeing up the database connection objects. """
        self._db_connection.close()
        self._db_connection = None

    def get_plugin_state(self, plugin_type, plugin_settings):
        """Return a dictionary of key-value pairs identifying that state of this plugin."""
        plugin_id = list(
            self._execute(
                "select plugin_id from plugin_registry where plugin_type = ? and settings = ?",
                (plugin_type, plugin_settings)
            )
        )

        if not plugin_id:
            raise PluginNotFoundError('Plugin not found in this index')

        else:
            plugin_state = self._execute(
                "select key, value from plugin_data where plugin_id = ?;",
                plugin_id[0]
            )

            for row in plugin_state:
                yield row

    def get_plugin_by_id(self, plugin_id):
        """Return the settings and state of the plugin identified by ID."""
        row = list(self._execute(
            'select plugin_type, settings from plugin_registry where plugin_id = ?', [plugin_id]
        ))
        if not row:
            raise PluginNotFoundError
        plugin_type, settings = row[0]
        state = self._execute("select key, value from plugin_data where plugin_id = ?", [plugin_id]).fetchall()
        return plugin_type, settings, state

    def list_known_plugins(self):
        """ Return a list of (plugin_type, settings, id) triples for each plugin stored in the index. """
        return list(self._execute("select plugin_type, settings, plugin_id from plugin_registry;"))

    @property
    def structured_fields(self):
        """Get a list of the structured field names on this index."""
        rows = list(self._execute('select name from structured_field'))
        return [row[0] for row in rows]

    @property
    def unstructured_fields(self):
        """Get a list of the unstructured field names on this index."""
        rows = list(self._execute('select name from unstructured_field'))
        return [row[0] for row in rows]

    def count_vocabulary(self, include_fields=None, exclude_fields=None):
        """Return the number of unique terms occuring in the given combinations of fields. """
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

        vocab_size = list(self._execute(
            'select count(distinct term_id) '
            'from term_statistics stats '
            'inner join unstructured_field field '
            '    on stats.field_id = field.id ' + where_clause,
            fields
        ))
        return vocab_size[0][0]

    def iterate_term_frequencies(self, terms=None, include_fields=None, exclude_fields=None):
        """Return a generator of frequencies over the list of terms supplied. """
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

        if terms is not None:
            term_filter = 'and voc.term in ({})'.format(', '.join(['?'] * len(terms)))
        else:
            term_filter = ''

        terms = terms or []

        frequencies = self._execute("""
            select voc.term, sum(frames_occuring)
            from term_statistics stats
            inner join vocabulary voc
               on voc.id = stats.term_id
               {}
            inner join unstructured_field field
               on stats.field_id = field.id
               {}
            group by voc.term
            """.format(term_filter, where_clause), terms + fields
        )
        return frequencies

    def _iterate_positions(self, terms=None, include_fields=None, exclude_fields=None):
        """Iterate through the positions index, giving frame ids and frequencies for matching terms.

        By default, all terms are iterated. Optionally a list of terms can be provided.

        """
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

        if terms is None:
            terms = (row[0] for row in self._execute(
                """
                select distinct term
                from term_statistics ts
                inner join vocabulary v
                    on ts.term_id = v.id
                inner join unstructured_field field
                    on field.id = ts.field_id
                {}
                """.format(where_clause),
                fields
            ))

        if fields:
            where_clause += 'and term = ?'
        else:
            where_clause = 'where term = ?'

        data = (fields + [term] for term in terms)
        frames = self._executemany(
            """
            select vocab.term, frame.id, field.name, post.frequency
            from term_posting post
            inner join vocabulary vocab
                on vocab.id = post.term_id
            inner join frame
                on frame.id = post.frame_id
            inner join unstructured_field field
                on field.id = frame.field_id
            {}
            order by term, frame.id
            """.format(where_clause),
            data
        )

        current_term = None

        for term, frame_id, field_name, frequency in frames:
            if current_term is None:
                positions = {frame_id: frequency}
                current_term = term
            elif term == current_term:
                positions[frame_id] = frequency
            else:
                yield current_term, positions
                positions = {frame_id: frequency}
                current_term = term
        else:
            if current_term is not None:
                yield current_term, positions

    def iterate_associations(self, term=None, association=None, include_fields=None, exclude_fields=None):
        """
        Term associations for this Index.

        This is used to record when two terms co-occur in a frame. Be aware that only 1 co-occurrence for two terms
        is recorded per frame no matter the frequency of each term. The format is as follows::

            {
                term: {
                    other_term: count,
                    ...
                },
                ...
            }

        Optionally the term or both the term and the association may be supplied.

        This method is a generator which yields a dict of {other_term: count, ...} for every term in the index.

        Note that this is dynamically calculated and may be expensive for large indexes.

        """

        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

        # If no field selectivity is defined, we can leave off the two joins:
        if fields:
            joined_where = """
                inner join frame
                    on frame.id = frame_post.frame_id
                inner join unstructured_field field
                    on field.id = frame.field_id
                {}""".format(where_clause)
        else:
            joined_where = ''

        # If a term is specified, add it to the join condition.
        term_filter = 'and left_vocab.term = ?' if term is not None else ''
        terms = [term] if term is not None else []
        association_filter = 'and right_vocab.term = ?' if association is not None else ''
        associations = [association] if association is not None else []

        rows = self._execute(
            """
            select
                left_vocab.term,
                right_vocab.term,
                count(*)
            from term_posting term_post
            inner join frame_posting frame_post
                on term_post.frame_id = frame_post.frame_id
                and term_post.term_id != frame_post.term_id
            inner join vocabulary left_vocab
                on left_vocab.id = term_post.term_id
                {}
            inner join vocabulary right_vocab
                on right_vocab.id = frame_post.term_id
                {}

            {}

            group by left_vocab.term, right_vocab.term
            order by left_vocab.term """.format(term_filter, association_filter, joined_where),
            terms + associations + fields
        )

        current_term = None

        for term, other_term, count in rows:
            if current_term is None:
                current_dict = {other_term: count}
                current_term = term

            elif current_term == term:
                current_dict[other_term] = count

            else:
                yield current_term, current_dict
                current_term = term
                current_dict = {other_term: count}

        else:  # Make sure to yield the final row.
            if current_term is not None:
                yield current_term, current_dict

    def count_documents(self):
        """Returns the number of documents in the index."""
        return list(self._execute('select count(*) from document'))[0][0]

    def count_frames(self, include_fields=None, exclude_fields=None):
        """Returns the number of documents in the index."""
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)
        return list(self._execute(
            'select count(*) from frame '
            'inner join unstructured_field field '
            '   on field.id = frame.field_id ' + where_clause,
            fields
        ))[0][0]

    def iterate_documents(self, document_ids=None):
        """Returns a generator  of (document_id, stored_document) pairs for the entire index.

        Optionally specify a list of document_ids to iterate over.

        The generator will only be valid as long as this reader is open.

        """
        if document_ids is not None:
            return self._executemany(
                'select * from document where id = ?',
                [[document_id] for document_id in document_ids]
            )
        else:
            return self._execute('select * from document')

    def iterate_frames(self, include_fields=None, exclude_fields=None, frame_ids=None):
        """Returns a generator  of (frame_id, document_id, field, sequence, stored_frame) tuples
        for the specified unstructured fields in the index.

        Optionally specify a list of frame_ids, in which case the field arguments will be ignored.

        The generator will only be valid as long as this reader is open.

        """
        if frame_ids is not None:
            return self._executemany(
                'select frame.id, document_id, field.name, sequence, stored '
                'from frame '
                'inner join unstructured_field field '
                '   on field.id = frame.field_id '
                'where frame.id = ?', [[frame_id] for frame_id in frame_ids]
            )
        else:
            where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)
            return self._execute(
                'select frame.id, document_id, field.name, sequence, stored '
                'from frame '
                'inner join unstructured_field field '
                '   on field.id = frame.field_id ' + where_clause,
                fields
            )

    def iterate_term_frequency_vectors(self, weighting='tf', include_fields=None, exclude_fields=None, frame_ids=None):
        """
        Iterates through sparse term_vectors for frames in the index.

        Currently only term frequency 'tf' weighting is supported.

        If frame_ids is provided, then the include_fields and exclude_fields arguments will be ignored.

        """

        if frame_ids is None:
            where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

            field_join = """
                inner join frame
                    on frame.id = frame_posting.frame_id
                inner join unstructured_field field
                    on field.id = frame.field_id
            """ if fields else ''
            rows = self._execute("""
                select frame_id, term, frequency
                from frame_posting
                inner join vocabulary
                    on frame_posting.term_id = vocabulary.id
                {}
                {}
                order by frame_id
            """.format(field_join, where_clause), fields)
        else:
            rows = self._executemany("""
                select frame_id, term, frequency
                from frame_posting
                inner join vocabulary
                    on frame_posting.term_id = vocabulary.id
                where frame_id = ?
            """, ((i,) for i in frame_ids))

        current_frame = None

        for frame_id, term, frequency in rows:
            if current_frame is None:
                term_freqs = {term: frequency}
                current_frame = frame_id

            elif current_frame == frame_id:
                term_freqs[term] = frequency

            else:
                yield current_frame, term_freqs
                current_frame = frame_id
                term_freqs = {term: frequency}

        else:  # Make sure to yield the final row.
            if current_frame is not None:
                yield current_frame, term_freqs

    def iterate_metadata(self, include_fields=None, exclude_fields=None, frames=True, text_field=None):
        """
        Get the metadata index.

        This method is a generator that yields tuples (field_name, value, [frame_ids])

        The frames flag indicates whether the return values should be broadcast to frame_ids (default)
        or document_ids.

        The optional text_field specifier allow filtering frames to just the included text field.

        """
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields, structured=True)

        if text_field and fields:
            where_clause += ' and text_field.name = ?'
            fields += [text_field]
        elif text_field:
            where_clause = ' where text_field.name = ?'
            fields += [text_field]

        if frames:
            rows = self._execute(
                'select field.name, value, frame.id '
                'from document_data '
                'inner join frame '
                '   on frame.document_id = document_data.document_id '
                'inner join structured_field field '
                '   on field.id = document_data.field_id '
                'inner join unstructured_field text_field '
                '   on text_field.id = frame.field_id ' + where_clause +
                'order by document_data.field_id, value',
                fields
            )
        else:
            rows = self._execute(
                'select field.name, value, document_id '
                'from document_data '
                'inner join structured_field field '
                '   on field.id = document_data.field_id ' + where_clause +
                'order by field_id, value',
                fields
            )
        current_field, current_value, document_id = next(rows)
        document_ids = [document_id]
        for row in rows:
            field, value, document_id = row
            # Rows are sorted by field, value, so as soon as the change can yield
            if field == current_field and value == current_value:
                document_ids.append(document_id)
            else:
                yield current_field, current_value, document_ids
                current_field = field
                current_value = value
                document_ids = [document_id]
        else:  # Make sure to yield the final row.
            yield current_field, current_value, document_ids

    def iterate_frame_attributes(self, frame_ids):
        """Iterate through the attributes of the given frames. """
        return self._executemany("""
            select
                type, value
            from frame_attribute_posting post
            inner join attribute
                on attribute.id = post.attribute_id
            where frame_id = ?
        """, [(f_id,) for f_id in frame_ids])

    def iterate_attributes(self, include_fields=None, exclude_fields=None, return_documents=False):
        """
        Get the frame-attribute index.

        This method is a generator that yields tuples (attribute_type, value, [frame_ids])

        The return_documents flag indicates whether the return values should be broadcast to frame_ids (default)
        or document_ids.

        """
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

        frames_or_documents = 'document_id' if return_documents else 'frame_id'

        if fields:
            frame_join = """
            inner join frame
                on frame.id = fp.frame_id
                and field_id in (select id from unstructured_field field {})
            """.format(where_clause)
        elif return_documents:
            frame_join = """
                inner join frame on frame.id = fp.frame_id
            """
        else:
            frame_join = ''

        query = """
            select {0}
                attribute.type,
                attribute.value,
                {1} as identifier
            from frame_attribute_posting fp
                {2}
            inner join attribute
               on attribute.id = fp.attribute_id
            order by type, value
        """.format('distinct' if return_documents else '', frames_or_documents, frame_join)

        rows = self._execute(query, fields)

        current_field, current_value, document_id = next(rows)
        document_ids = [document_id]

        for row in rows:
            field, value, document_id = row
            # Rows are sorted by field, value, so as soon as the change can yield
            if field == current_field and value == current_value:
                document_ids.append(document_id)
            else:
                yield current_field, current_value, document_ids
                current_field = field
                current_value = value
                document_ids = [document_id]
        else:  # Make sure to yield the final row.
            yield current_field, current_value, document_ids

    def iterate_bigram_positions(self, bigrams, include_fields=None, exclude_fields=None):
        """Return an iterator of (left_term, right_term, frame_id, frequency) tuples for the specified list of bigrams.

        Bigrams are supplied as a list of tuples: [('apple', 'pie'), ('whipped', 'cream')].

        Currently, only exact matches for each term are considered - if one of the terms in the bigram occurs
        after the 63rd position in a frame it is not considered a match.

        """
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

        if fields:
            extra_join = """
                inner join frame
                    on frame.id = right_post.frame_id
                inner join unstructured_field field
                    on field.id = frame.field_id
            """
            # Mangle the normal where clause for fielded search: this query is difficult to structure to match this.
            extra_where = 'and ' + where_clause[5:]
        else:
            extra_where = extra_join = ''

        query_data = (list(bigram) + fields for bigram in bigrams)
        bigrams = self._executemany("""
            select
                left_vocab.term,
                right_vocab.term,
                left_post.frame_id,
                left_post.positions & (right_post.positions >> 1) as matched_positions
            from term_posting left_post
            inner join term_posting right_post
                on left_post.frame_id = right_post.frame_id
            inner join vocabulary left_vocab
                on left_vocab.id = left_post.term_id
                and left_vocab.term = ?
            inner join vocabulary right_vocab
                on right_vocab.id = right_post.term_id
                and right_vocab.term = ?
            {}
            where
                -- Exclude approximate positions indexes
                left_post.positions > 0
                and right_post.positions > 0
                -- And they actually have matching positions
                and matched_positions > 0
                {}
        """.format(extra_join, extra_where), query_data
        )

        for left_term, right_term, frame_id, positions in bigrams:
            yield ((left_term, right_term), frame_id, _count_bitwise_matches(positions))

    def filter_range(
        self, start, end=None, limit=None, return_documents=False, include_fields=None, exclude_fields=None
    ):
        """
        Return frame or document ID's greate than start up to and including an optional end key, up to the limit
        supplied. If there are fewer than limit frames or documents, all remaining id's will be returned.

        The include_fields and exclude_fields parameters are ignored when return_documents is True.

        The range interval is open on the left - you can use the end of one range as the start parameter
        to page through the results.

        """
        query_template = """
            select id
            from {0}
            where {1}
            {2}
        """

        parameters = []

        if return_documents:
            frame_or_document = 'document'

        else:
            where_clause, fields = self._fielded_where_clause(
                include_fields=include_fields, exclude_fields=exclude_fields
            )
            if fields:
                frame_or_document = """
                (select id from frame
                 where frame.field_id in (
                    select id from unstructured_field field
                    {}
                ))
                """.format(where_clause)
                parameters.extend(fields)
            else:
                frame_or_document = 'frame'

        id_clause = 'id > ? ' if end is None else 'id > ? and id <= ?'
        parameters.extend([start] if end is None else [start, end])

        if limit is not None:
            limit_clause = 'limit ?'
            parameters.append(limit)
        else:
            limit_clause = ''

        query = query_template.format(frame_or_document, id_clause, limit_clause)

        return self._execute(query, parameters)

    def filter_metadata(
        self, metadata, return_documents=False, include_fields=None, exclude_fields=None,
        limit=0, pagination_key=None
    ):
        """
        Support metadata only searches - for efficiency reasons filter_or_rank_unstructured
        is driven by the term_posting table. This function provides efficient set filtering for
        metadata only cases.

        Currently only conjunctive metadata queries are supported, with the exception of the 'in'
        operator for matching multiple alternatives.

        The include_field and exclude_field arguments refer to unstructured fields only - these
        options are ignored if return_documents is True.

        See rank_or_filter_unstructured for more detail.

        """

        metadata_clauses = []
        parameters = []
        # Whitelist of valid operators - we can't bind an operator in sql, so reject everything else.
        valid_metadata_operators = set(('<', '>', '<=', '>=', '=', 'in'))

        for metadata_field, operators in metadata.items():

            # Template for each field - fill in the field name and append the parameters.
            this_field = """
                select document_id from document_data
                where field_id = (select id from structured_field where name = ?)
            """

            parameters.append(metadata_field)

            if return_documents and pagination_key:
                this_field += 'and document_id > ? '
                parameters.append(pagination_key)

            for operator, value in operators.items():
                if operator not in valid_metadata_operators:
                    raise ValueError('{} is not a supported operator for SQLiteStorage'.format(operator))

                if operator == 'in':  # The values for an 'in' operator should be an iterable.
                    this_field += 'and value {} ({}) \n'.format(operator, ', '.join(['?'] * len(value)))
                    parameters.extend(value)

                else:
                    this_field += 'and value {} ? \n'.format(operator)
                    parameters.append(value)

            metadata_clauses.append(this_field)

        # Note that all metadata queries are conjunctions only: all metadata clauses must be matched.
        document_intersection = '\n intersect \n'.join(metadata_clauses)

        if return_documents:
            if limit:
                document_intersection += 'order by document_id \n limit ?'
                parameters.append(limit)

            results = self._execute(document_intersection, parameters)

        else:
            frame_intersection = """
                select id
                from frame
                where document_id in (
                    {}
                )
                {}
                {}
                {}
                """
            where_field, fields = self._fielded_where_clause(include_fields, exclude_fields)
            if fields:
                field_selector = 'and field_id in (select field.id from unstructured_field field {})'
                field_selector = field_selector.format(where_field)
                parameters.extend(fields)
            else:
                field_selector = ''
            if pagination_key:
                pagination_clause = 'and frame.id > ?'
                parameters.append(pagination_key)
            else:
                pagination_clause = ''
            if limit:
                limit_clause = 'order by frame.id \n limit ?'
                parameters.append(limit)
            else:
                limit_clause = ''

            results = self._execute(
                frame_intersection.format(document_intersection, field_selector, pagination_clause, limit_clause),
                parameters
            )

        return results

    def rank_or_filter_unstructured(
        self, return_documents=False, include_fields=None, exclude_fields=None,
        must=None, should=None, at_least_n=None, must_not=None, metadata=None,
        limit=0, pagination_key=None, search=False
    ):
        """
        Omnibus function for searches and filters on unstructured data.

        This function provides a fast low level search query interface, suitable for producing
        fast resultsets for combining with caterpillar.combination functions.

        Currently only conjunctive metadata searches are supported.

        The cost of a search is driven by the cost of the largest intermediate term set:
        a search in a large English language index for the term 'and' will be expensive,
        regardless of which criteria (must, should etc.) is used as all rows containing that
        term will need to aggregated over.

        Scoring is currently based on TF-IDF. Frames are scored individually. Document scores are
        computed by summing frame scores. The inverse 'document' frequency is actually the
        inverse frame frequency of that term across the included unstructured fields in this
        query.

        For filtering, the pagination key is the last frame_id or document_id seen, for search,
        the pagination key is a tuple of (frame_id, score). Pagination based on a deterministic
        key gives better performance on deep paging queries (later queries are slightly cheaper
        rather than more and more expensive), and also gives a more stable sort with respect to
        relevance when new results occur.

        Current limitations:
            - must_not is not supported without must/should or at_least_n
            - wildcards (*/?) are not supported
            - Arbitrary query trees are not supported - this provides a 'flat' fast interface
              only.

        See also:
            filter_metadata: path for metadata only queries.

        """

        # This is the query we will be passing through to SQLite. The remainder of this function
        # just fills in the gaps, conditioned on all the options specified by the user.
        query = """
            select
                {frame_or_document},
                sum(frequency * ts.weight) as score
            from term_search_driver ts

            /* Optimisation Note

            Note that in SQLite cross join behaves identically to inner join, *except* the query
            optimiser does not reorder the joins. The join order is critical for this case - if
            term_posting is reordered we may end up with a full table scan of a very large table.

            */

            cross join term_posting post
                on ts.term_id = post.term_id

            {subset_clause}

            {filter_pagination}

            {groupby}

            having 1 -- no-op clause to simplify query construction.
                {having}

            {search_clause}
        """

        # Parameter handling, expand optional parameters for the rest of the function
        must = must or []
        should = should or []
        at_least_n = at_least_n or (0, [])
        must_not = must_not or []
        metadata = metadata or {}

        # Raise an error if must_not is specified without should, must or at_least_n
        if must_not and not(at_least_n[1] or should or must):
            raise ValueError(
                '"must_not" is not supported for SQLiteStorage without at least one term in '
                '"must", "should" or "at_least_n"'
            )

        if search and not(at_least_n[1] or should or must):
            raise ValueError(
                'Ranking is not supported for metadata only searches in SQLiteStorage'
            )

        # Expand singular terms into lists for each variant to support variant search.
        # ['term1', 'term2', ('term3', 'term4')] --> [('term1'), ('term2'), ('term3', 'term4')]
        # Each tuple of terms is assigned a single search_id for determining matches.
        # Note that for must_not and should the variation syntax is supported, even if there
        # is no difference in behaviour.
        search_term_groups = [_unpack_mixed_term_list(group) for group in [must, at_least_n[1], must_not, should]]
        search_terms = [[term] for group in search_term_groups for terms in group for term in terms]

        # If there are no terms or metadata specified, there are no results as this search
        # is driven by the matching terms in must, should and at_least_n.
        if len(search_terms) == 0 and not metadata:
            return [None]

        # Construct one row per term in the search to insert in the driving table.
        # Note that grouped terms get assigned the same search_id - this is magic to make it work.
        search_id_rows = []
        search_id = 1

        for i, group in enumerate(search_term_groups):
            for terms in group:
                for term in terms:
                    # Columns: all_id, n_id, exclude_count, (not included in table)
                    search_row = [None, None, 0, None]
                    search_row[i] = search_id
                    search_id_rows.append(search_row)
                search_id += 1

        # Generate the where clause, including the metadata specific details.
        unstructured_where_clause, unstructured_fields = self._fielded_where_clause(include_fields, exclude_fields)

        # Compute IDF component of the term weighting from the term_statistics on this index
        n_frames = list(self._execute(
            'select sum(frame_count) '
            'from field_statistics '
            'inner join unstructured_field field on field_statistics.field_id = field.id ' +
            unstructured_where_clause, unstructured_fields)
        )[0][0]

        # Note here - this returns Null if a term doesn't exist
        term_stats = list(self._executemany(
            """
            select sum(frames_occuring) as frame_frequency
            from term_statistics
            inner join vocabulary
                on term_statistics.term_id = vocabulary.id
            where vocabulary.term = ?
            """, search_terms)
        )

        # Early exit if none of the terms match.
        if sum(1 for i in term_stats if i[0] is not None) == 0:
            if search:
                return []
            else:
                return {}

        # The none branch handles if the term lookup failed
        term_idf = [(1 + math.log(n_frames / (n[0] + 1))) if n[0] is not None else 0 for n in term_stats]

        # Truncate the temporary driving table
        # Note that because of how searches work, this means that only a single search query
        # can be active for a given reader, as SQLite temporary tables are only isolated across connections.
        # For this reason, although this method can potentially be used as a generator,
        # the IndexReader API always returns the complete resultset.
        self._execute('delete from term_search_driver')

        # Stage the terms to the driving table, including the necessary weighting
        self._executemany("""
            insert into term_search_driver(term_id, all_id, n_id, exclude_count, weight)
                select id as term_id, ?2, ?3, ?4, ?5
                from vocabulary
                where term = ?1
                order by term_id
            """, [term + row[:3] + [weight] for term, row, weight in zip(search_terms, search_id_rows, term_idf)]
        )

        parameters = []

        if metadata or unstructured_fields or return_documents:
            subset_clause = 'inner join frame on post.frame_id = frame.id '

            if unstructured_fields:
                subset_clause += ' and frame.field_id in (select id from unstructured_field field {})'.format(
                    unstructured_where_clause
                )
                parameters += unstructured_fields

            # Note that by this point, all of the metadata values must be analysed and the operators validated by the
            # IndexReader. In this storage layer we are dealing only with the representation of the value in the
            # database.
            if metadata:

                metadata_clauses = []
                # Whitelist of valid operators - we can't bind an operator in sql, so reject everything else.
                valid_metadata_operators = set(('<', '>', '<=', '>=', '=', 'in'))

                for metadata_field, operators in metadata.items():

                    this_field = (
                        'select document_id from document_data '
                        'where field_id = (select id from structured_field where name = ?)'
                    )
                    parameters.append(metadata_field)

                    for operator, value in operators.items():
                        if operator not in valid_metadata_operators:
                            raise ValueError('{} is not a supported operator'.format(operator))

                        if operator == 'in':  # The values for an 'in' operator should be an iterable.
                            this_field += 'and value {} ({}) \n'.format(operator, ', '.join(['?'] * len(value)))
                            parameters.extend(value)

                        else:
                            this_field += 'and value {} ? \n'.format(operator)
                            parameters.append(value)

                    metadata_clauses.append(this_field)

                # Note that all metadata queries are intersections: all metadata clauses must be matched.
                subset_clause += 'and document_id in ({})'.format(' intersect '.join(metadata_clauses))

        else:
            subset_clause = ''

        if not search and return_documents and pagination_key is not None:
            filter_pagination = 'where document_id > ?'
            parameters.append(pagination_key)
        elif not search and pagination_key is not None:
            filter_pagination = 'where frame_id > ?'
            parameters.append(pagination_key)
        else:
            filter_pagination = ''

        if return_documents:
            groupby_clause = 'group by document_id'
        else:
            groupby_clause = 'group by frame_id'

        having_clause = ''
        # Construct term_inclusion clauses: avoid checking things that aren't needed.
        if must:
            # if all_count is NULL, term || all_count evaluates to null, and is not counted in the distinct
            having_clause += 'and count(distinct all_id) = ? '
            parameters.append(len(must))

        if at_least_n[0]:
            having_clause += 'and count(distinct n_id) >= ? '
            parameters.append(at_least_n[0])

        if must_not:
            having_clause += 'and max(exclude_count) = 0 '

        search_clause = ''

        if search:
            if pagination_key is not None:
                having_clause += 'and (score < ? or (score = ? and {} > ?))'.format(
                    'document_id' if return_documents else 'frame_id'
                )
                parameters.extend([pagination_key[1], pagination_key[1], pagination_key[0]])

            # If we're searching, order by score descending and frame/document_id ascending for deterministic pagination
            search_clause += 'order by score desc, {} '.format('document_id' if return_documents else 'frame_id')

        elif limit:
            search_clause += 'order by {} '.format('document_id' if return_documents else 'frame_id ')

        if limit:
            search_clause += 'limit ?'
            parameters.append(limit)

        # Now fill in the template, and actually execute the query.
        results = self._execute(
            query.format(
                frame_or_document='document_id' if return_documents else 'frame_id',
                filter_pagination=filter_pagination,
                subset_clause=subset_clause,
                groupby=groupby_clause,
                having=having_clause,
                search_clause=search_clause
            ), parameters
        )

        return results

    def filter_attributes(
        self, attributes, return_documents=False, include_fields=None, exclude_fields=None, limit=0, pagination_key=None
    ):
        """
        Return frames or documents containing specific attributes.

        Currently this is a very thin skin over the underlying tables - expect this interface to
        be merged with regular term filtering and the schema field API in the future. No type conversion or checking
        is performed - the attribute comparisons will be directly used as bound parameters in an SQL query.

        Don't rely on this API to be stable!

        """
        operator_whitelist = {'=', '>', '>=', '<', '<='}
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)
        parameters = []

        if fields:
            field_clause = """
                inner join frame
                    on afp.frame_id = frame.id
                    and field_id in (select id from unstructured_field field {})
            """.format(where_clause)
        elif return_documents:
            field_clause = """
                inner join frame
                    on afp.frame_id = frame.id
            """
        else:
            field_clause = ''

        attribute_block = """
            select {0}
            from attribute
            cross join attribute_frame_posting afp
                on afp.attribute_id = attribute.id
                {1}
            where
                attribute.type = ?
                {2}
                {3}
        """

        documents_or_frames = 'document_id' if return_documents else 'frame_id'
        pagination_clause = 'and {} > ?'.format(documents_or_frames) if pagination_key is not None else ' '

        attribute_blocks = []

        for attribute_type, value_comparisons in attributes.items():
            if any(operator not in operator_whitelist for operator in value_comparisons.keys()):
                raise ValueError('Only {} operators are supported.'.format(operator_whitelist))

            op_values = value_comparisons.items()
            operator_clauses = '\n'.join(['and value {} ?'.format(op) for op, _ in op_values])
            parameters.append(attribute_type)
            parameters.extend(fields)
            parameters.extend([v for _, v in op_values])
            if pagination_key is not None:
                parameters.append(pagination_key)

            attribute_blocks.append(
                attribute_block.format(
                    documents_or_frames, field_clause, operator_clauses, pagination_clause)
            )

        if limit:
            limit_clause = 'limit ?'
            parameters.append(limit)
        else:
            limit_clause = ''

        full_query = '{0} {1}'.format("""
            intersect
        """.join(attribute_blocks), limit_clause
        )

        return self._execute(full_query, parameters)

    def find_significant_bigrams(self, include_fields=None, exclude_fields=None, min_count=5, threshold=40):
        """Find significant collocations of words.

        Currently operates over all fields in the index.

        Currently, only exact matches for each term are considered - if one of the terms in the bigram occurs
        after the 63rd position in a frame it is not considered a match.

        Algorithm Notes

        The formula for calculating bi-gram score is inspired by the Gensim implementation of phrase detection from the
        Mikolov et al paper, "Distributed Representations of Words and Phrases and their Compositionality".

        score(a, b) = freq(a, b) * vocab_size / (freq(a) * freq(b))

        Currently the frequencies are the number of frames a bigram/unigram occurs in.

        Args

            min_count: specifies the minimum number of times a bigram must occur to be considered. It is also
                used to prefilter the vocabulary for terms that don't occur enough to form a bigram.
            threshold: the value of the statistical threshold used to determine if a phrase is a match or not.

        """
        where_clause, fields = self._fielded_where_clause(include_fields, exclude_fields)

        # If fields are specified, we have some extra work to do.
        if fields:
            post_join = """
                inner join frame
                    on frame.id = right_post.frame_id
                inner join unstructured_field field
                    on field.id = frame.field_id
            """
            # Mangle the normal where clause for fielded search: this query is difficult to structure to match this.
            post_where = 'and ' + where_clause[5:]
            term_join = 'inner join unstructured_field field on field.id = ts.field_id and ' + where_clause[5:]
        else:
            post_where = post_join = term_join = ''

        bigrams = self._execute("""
            with bigrams as (
                select
                    left_post.term_id as left_id,
                    right_post.term_id as right_id,
                    count(*) * 1.0 as bigram_count
                from frame_posting left_post
                inner join frame_posting right_post
                    on left_post.frame_id = right_post.frame_id
                {}
                where
                    -- Exclude approximate positions indexes
                    left_post.positions > 0
                    and right_post.positions > 0
                    -- And they actually have matching positions
                    and (left_post.positions & (right_post.positions >> 1)) > 0
                    {}
                group by left_post.term_id, right_post.term_id
                having bigram_count > ?
            ),
            field_statistics as (
                select ts.term_id, term, sum(frames_occuring) as frames_occuring
                from term_statistics ts
                inner join vocabulary
                    on vocabulary.id = ts.term_id
                {}
                group by ts.term_id, term
            )
            select left_stats.term, right_stats.term
            from bigrams
            inner join field_statistics left_stats
                on left_stats.term_id = bigrams.left_id
            inner join field_statistics right_stats
                on right_stats.term_id = bigrams.right_id
            where (
                bigram_count * (select count(*) from field_statistics) /
                (1.0 * left_stats.frames_occuring * right_stats.frames_occuring)
            ) > ?
            """.format(post_join, post_where, term_join),
            fields + [min_count] + fields + [threshold]
        )

        return bigrams

    def get_settings(self, names):
        """Get the settings identified by the given names. """
        variable_binding = ', '.join(['?'] * len(names))
        return self._execute(
            'select * from setting where name in ({})'.format(variable_binding),
            names
        )

    @property
    def revision(self):
        """
        The revision identifier is a tuple (revision_number, added_documents, deleted_documents)

        The revision number is incremented on every writer commit, the added documents and deleted
        documents count the number of times add_analyzed_document and delete_document were
        succesfully called on a writer for this index.
        """
        revision = list(self._execute(
            'select * from index_revision where revision_number=(select max(revision_number) from index_revision)'
        ))
        return revision[0]

    def _execute(self, query, data=None):
        cursor = self._db_connection.cursor()
        try:
            return cursor.execute(query, data)
        except apsw.SQLError as e:
            logger.exception(e)
            raise e

    def _executemany(self, query, data=None):
        cursor = self._db_connection.cursor()
        try:
            return cursor.executemany(query, data)
        except apsw.SQLError as e:
            logger.exception(e)
            raise e

    def _fielded_where_clause(self, include_fields, exclude_fields, structured=False):
        """Generate a where clause for field inclusion, validating the fields at the same time.

        Include fields takes priority if both include and exclude fields are specified.

        Returns both the where clause and the list of fields to filter on for binding.

        """
        fields = include_fields or exclude_fields or []
        valid_fields = self.structured_fields if structured else self.unstructured_fields
        # Catch None as a valid field to allow current reader level interface to specify None as a field.
        invalid_fields = [field for field in fields if field not in valid_fields and field is not None]

        if invalid_fields:
            raise NonIndexedFieldError('Invalid fields: {} do not exist or are not indexed'.format(invalid_fields))
        if include_fields:
            where_clause = 'where field.name in ({})'.format(', '.join(['?'] * len(include_fields)))
        elif exclude_fields:
            where_clause = 'where field.name not in ({})'.format(', '.join(['?'] * len(exclude_fields)))
        else:
            where_clause = ''
        return where_clause, fields


SqliteStorage = Storage(SqliteReader, SqliteWriter)


def _bitwise_encode(ordinal_positions):
    """
    Converts the sorted list of integers to a bitstring.

    The integer i is indicated in the positions by setting the i'th bit of the string to 1.

    Superpositions of integers up to 62 (positions 0, 1, ... 62) can be represented exactly.

    For integers larger than 62, an approximate matching scheme is used: the position i % 63
    is recorded instead. If the match is approximate, the high bit will set: the output integer is
    negative if the match is approximate.

    """

    p = 0

    for i in ordinal_positions:
        p |= 1 << i % 63

    if i > 62:
        p = -p

    return p


def _count_bitwise_matches(position_bitstring):
    """ Count the number of matches (number of bits set to 1) indicated by the given bitstring.

    If the high bit is set in the bitstring, this will return 0.

    This uses Kernighan's algorithm, and is faster in the case of a small number of bits set.

    """
    if position_bitstring <= 0:
        return 0

    n = 0

    while position_bitstring > 0:
        position_bitstring &= position_bitstring - 1
        n += 1

    return n


def _unpack_mixed_term_list(term_sequence):
    """ Unpack the list of terms into term groups.

    Assume that anything this isn't a string is an iterable that can be left along.

    """
    return [
        [group] if isinstance(group, basestring) else group for group in term_sequence
    ]



# Copyright (c) Kapiche Limited
# Author: Sam Hames <sam.hames@kapiche.com>
"""
The schema scripts for the bulk operations of the :class:`.SqliteStorage`.

"""

disk_schema = """
/* pragma note: the page_size is set for the whole database, and can only change with
and expensive vaccuum operation - it also must be run before anything else is done
with the database. So this line must come first! */
pragma page_size = 4096; -- current recommended value for SQLite.

/* Set write ahead log mode for all connections that will use this database. The WAL
allow concurrent readers and writers, unlike the default journal mode that locks readers
out while writing. */
pragma journal_mode = WAL;

begin;

/* Field names and ID's

Structured and unstructured fields are kept separate because the storage and querying
of each type of data is different.
*/
create table structured_field (
    id integer primary key,
    name text unique
);

create table unstructured_field (
    id integer primary key,
    name text unique
);


/*
The core vocabulary table assigns an integer ID to every unique term in the index.

Joins against the core posting tables are always integer-integer and in sorted order.

Currently this table allows repeating terms to support destructive merge variants and
case folding. In the future this table will become append only and term uniqueness will
be enforced, and an alternate mechanism for specifying vocabulary variations will be provided.
*/
create table vocabulary (
    id integer primary key,
    term text
);
create index term_idx on vocabulary(term);


/* Summary statistics for a given term_id by field.

This allows direct lookups for Tf.IDF searches and similar.
*/
create table term_statistics (
    term_id integer,
    field_id integer,
    frequency integer,
    frames_occuring integer,
    documents_occuring integer,
    primary key(term_id, field_id),
    foreign key(term_id) references term(id),
    foreign key(field_id) references field(id)
);


/* Summary statistics for each field. */
create table field_statistics (
    field_id integer primary key,
    frame_count integer
);


/* The source table for the document representation. */
create table document (
    id integer primary key,
    stored text -- This should be a text serialised representation of the document, such as JSON.
);

/*
Storage for 'indexed' structured fields in the schema.

- designed for sparse data and extensible schema's
- primary design purpose is for returning ordered lists of document ID's
- takes advantage of SQLite's permissive type system

*/
create table document_data (
    field_id integer,
    value,
    document_id integer,
    primary key(field_id, value, document_id),
    foreign key(document_id) references document(id),
    foreign key(field_id) references structured_field(id)
);


/* Storage for frames, structured to link between document level metadata and frame level text information. */
create table frame (
    id integer primary key,
    document_id integer,
    field_id integer,
    sequence integer,
    stored text -- The stored representation of the frame.
);


/* Index to access by document ID

Bridges between:
structured data searches --> frames
unstructured searches --> documents
*/
create index document_frame_bridge on frame(document_id, field_id);
create index field_frame_idx on frame(field_id, document_id);


/* Postings organised by term, allowing search operations. */
create table term_posting (
    term_id integer,
    frame_id integer,
    frequency integer,
    positions integer,
    primary key(term_id, frame_id),
    foreign key(term_id) references term(id),
    foreign key(frame_id) references frame(id)
)
without rowid; -- Ensures that the data in the base table is kept in this sorted order.

/* Postings organised by frame, for term-frequency vector representations of documents and frames. */
create table frame_posting (
    frame_id integer,
    term_id integer,
    frequency integer,
    positions integer,
    primary key(frame_id, term_id),
    foreign key(term_id) references term(id) on delete cascade
    foreign key(frame_id) references frame(id) on delete cascade
)
without rowid;


/* Frame attribute storage.

This is the equivalent of the vocabulary table for semi-structured frame level data. This
table allows a 1:many mapping of frames --> attributes. An attribute is any structured piece
of information that can be extracted from a frame, such as sentiment scores, named entities,
parsed email addresses or websites, twitter @handles and #hashtags etc.

Although the data layout is conceptually similar to the term_posting + vocabulary table
it is maintained separately here for two reasons:
    1. Avoid overcomplicating the text search API with the addition of attribute types. (Also
       avoid rewriting the complex queries in search.)
    2. Simpler API - in the short term it is expected that attributes will be added
       post indexing/separate to plain text indexing. This means it is advantageous to have
       the attributes separate to the main table to better enhance write locality.

In the future we may consider merging the two tables and having a unified view into
unstructured attributes of a dataset. This might also allow more nuanced interpretation
of text indexing - for example indexing terms with an attribute class of their parts of
speech to allow more complex search operators/ disambiguation.

*/

/* Actual storage of the values of attributes.

Note - if an attribute class should be 1:1 on a frame, this needs to be enforced at the
application/reader level, as this schema needs to necessarily accept 1:many relationships.

*/
create table attribute (
    id integer primary key,
    type text,
    value,
    constraint attr_class_value unique (type, value) on conflict ignore
);


/* Attribute-frame tables

Analogous to frame_posting and term_posting, store attribute presence on a
frame, ordered by frame for analytical operations and by attribute for search operations.

*/
create table frame_attribute_posting (
    frame_id integer,
    attribute_id integer,
    primary key (frame_id, attribute_id) on conflict ignore,
    foreign key (frame_id) references frame(id),
    foreign key (attribute_id) references attribute(id)
)
without rowid;

create table attribute_frame_posting (
    attribute_id integer,
    frame_id integer,
    primary key (attribute_id, frame_id) on conflict ignore,
    foreign key (frame_id) references frame(id),
    foreign key (attribute_id) references attribute(id)
)
without rowid;


/* Plugin header and data tables. */
create table plugin_registry (
    plugin_type text,
    settings text,
    plugin_id integer primary key,
    constraint unique_plugin unique (plugin_type, settings) on conflict ignore
);

create table plugin_data (
    plugin_id integer,
    key text,
    value text,
    primary key(plugin_id, key) on conflict replace,
    foreign key(plugin_id) references plugin_registry(plugin_id) on delete cascade
);


/*
An internal representation of the state of the index documents.

The state of the documents in the index is represented by a tuple of:
    (revision_number, added_document_count, deleted_document_count, added_frame_count)

Each number is monotonically increasing and represents the current state of the index over it's
lifetime. The next document and frame added to the index will have ID's of added_document_count + 1
and added_frame_count + 1.

The revision number is incremented by one whenever a write transaction adds or deletes documents
from an index.

For example, a reader might run an operation when the state of the index was (5, 100, 5, 2000), while
the current state is (15, 130, 45, 2200). This means that 30 new documents and 200 new frames were added
in 10 commits, and 40 documents were deleted since that operation was run.

*/
create table index_revision (
    revision_number integer primary key,
    added_document_count integer,
    deleted_document_count integer,
    added_frame_count integer,
    constraint unique_revision unique(
        added_document_count, deleted_document_count, added_frame_count
    ) on conflict ignore
);

insert into index_revision values (0, 0, 0, 0);


create table setting (
    name text primary key on conflict replace,
    value
);


commit;

"""

"""

/*
A convenience view to simplify writing search queries.

If we move to a segmented or otherwise optimised index structure this view will
combine the tables, so queries should use this in preference to direct table references.

Note that this view uses the canonical representation of the term to represent variants.
*/

create view search_posting as (
    select *
    from term_posting
    inner join vocabulary
        on term_posting.term_id = vocabulary.id
    inner join frame
        on term_posting.frame_id = frame.id
    inner join field
        on frame.field_id = field.id
    where document_id not in (select document_id from deleted_documents)
);
"""

# Schema for the staging database
cache_schema = """
begin;

create table structured_field (
    name text primary key
);
create table unstructured_field (
    name text primary key
);

/* The source table for the document representation. */
create table document (
    id integer primary key,
    stored text -- This should be a text serialised representation of the document, such as JSON.
);

create table deleted_document(
    id integer primary key
);

/* Storage for 'indexed' structured fields in the schema. */
create table document_data (
    document_id integer,
    field_name text,
    value,
    primary key(document_id, field_name)
);

create table frame (
    id integer primary key,
    document_id integer,
    field_name text,
    sequence integer, -- The sequence number of the frame in that field of the document
    stored text -- The stored representation of the frame
);

/* One row per term occuring in a frame */
create table stage_posting (
    frame_id integer,
    term text,
    frequency integer,
    positions text,
    primary key(frame_id, term)
);

/* one row per attribute in a frame. */
create table attribute_posting (
    frame_id integer,
    type text,
    value,
    primary key (frame_id, type, value)
);

create table setting(
    name text primary key on conflict replace,
    value
);

/* Plugin header and data tables. */
create table plugin_registry (
    plugin_type text,
    settings text,
    constraint unique_plugin unique (plugin_type, settings) on conflict ignore
);

create table plugin_data (
    plugin_type text,
    settings text,
    key text,
    value text,
    primary key(plugin_type, settings, key) on conflict replace
);

create table delete_plugin (
    plugin_type text,
    settings text
);


commit;
begin immediate;

"""

# Prepare commit by precalculating and sorting everything.
# Returns rows representing the current maximum frame and document ID's for assigning
# ID's during the script execution.
prepare_flush = """
-- For generating the term-frame_id and attribute-frame_id tables
create index term_idx on stage_posting(term, frame_id);
create index attribute_idx on attribute_posting(type, value, frame_id);
create table distinct_attributes as
select distinct type, value from attribute_posting;

-- Generate term statistics of added documents
create table term_statistics as
    select
        term,
        field_name,
        sum(frequency) as frequency,
        count(distinct frame_id) as frames_occuring,
        count(distinct document_id) as documents_occuring
    from stage_posting pos
    inner join frame
        on frame_id = frame.id
    group by
        pos.term,
        frame.field_name;

create index term_stats_idx on term_statistics(term, field_name);

commit; -- end staging transaction so we can attach on disk database.

-- Attach the on-disk database to flush to.
attach database ? as disk_index;

begin immediate; -- Begin the true transaction for on disk writing

-- Max document and frame id's at the start of the write process.
select * from index_revision
where revision_number = (select max(revision_number) from index_revision);

-- Actual ID's of deleted documents, for reporting succesful deletion and updating delete counts
select distinct id from deleted_document where id in (select id from disk_index.document);

"""

# Flush changes from the cache to the index
flush_cache = """
/* Delete documents */
-- Cache term statistics for deleted documents, note the negative for later.
create table deleted_term_statistics as
    select
        term_id,
        field_id,
        -sum(frequency) as frequency,
        -count(distinct frame_id) as frames_occuring,
        -count(distinct document_id) as documents_occuring
    from disk_index.frame_posting post
    inner join disk_index.frame frame
        on frame_id = frame.id
    where frame.document_id in (select * from deleted_document)
    group by
        post.term_id,
        frame.field_id;

create table deleted_frame as
    select id
    from disk_index.frame
    where document_id in (select * from deleted_document);
create index deleted_frame_idx on deleted_frame(id);

delete from disk_index.term_posting
-- Avoid full table scan by searching for term_id first
-- This is still going to be expensive.
where term_id in (select distinct term_id
                  from disk_index.frame_posting
                  where frame_id in (select * from deleted_frame))
    and frame_id in (select * from deleted_frame);

delete from disk_index.frame_posting
    where frame_id in (select * from deleted_frame);

delete from disk_index.attribute_frame_posting
-- Avoid full table scan by searching for attribute_id first
-- This is still going to be expensive.
where attribute_id in (
    select distinct attribute_id
    from disk_index.frame_attribute_posting
    where frame_id in (select * from deleted_frame)
)
    and frame_id in (select * from deleted_frame);

delete from disk_index.frame_attribute_posting
    where frame_id in (select * from deleted_frame);

-- Delete all the places the document occurs.
delete from disk_index.document where id in (select * from deleted_document);
delete from disk_index.frame where document_id in (select * from deleted_document);
delete from disk_index.document_data where document_id in (select * from deleted_document);


/* Add new indexed fields */
insert into disk_index.structured_field(name)
    select * from structured_field;
insert into disk_index.unstructured_field(name)
    select * from unstructured_field;


/* Update vocabulary with new terms. Insert highest frequency first. */
insert into disk_index.vocabulary(term)
    select term
    from term_statistics stats
    where not exists (
        select 1
        from vocabulary v
        where v.term = stats.term
    )
    group by term
    order by sum(frequency) desc;


/* Insert document and frame data */
insert into disk_index.document(id, stored)
    select
        id + :max_doc,
        stored
    from document;


insert into disk_index.document_data(document_id, field_id, value)
    select
        document_id + :max_doc,
        fields.id,
        value
    from document_data data
    inner join disk_index.structured_field fields
        on fields.name = data.field_name;


insert into disk_index.frame(id, document_id, field_id, sequence, stored)
    select
        frame.id + :max_frame,
        document_id + :max_doc,
        fields.id,
        sequence,
        stored
    from frame
    inner join disk_index.unstructured_field fields
        on fields.name = frame.field_name;


/* Term and frame posting data */
insert into disk_index.frame_posting(frame_id, term_id, frequency, positions)
    select
        pos.frame_id + :max_frame,
        vocab.id,
        frequency,
        positions
    from stage_posting pos
    inner join disk_index.vocabulary vocab
        on vocab.term = pos.term;


insert into disk_index.term_posting(term_id, frame_id, frequency, positions)
    select
        vocab.id,
        pos.frame_id + :max_frame,
        frequency,
        positions
    from stage_posting pos
    inner join disk_index.vocabulary vocab
        on vocab.term = pos.term
    order by vocab.id;


/* Insert new attribute-value pairs */
insert into disk_index.attribute(type, value)
    select type, value
    from distinct_attributes attr
    where not exists (
        select 1
        from disk_index.attribute at
        where attr.type = at.type
            and attr.value = at.value
    );

/* Add to the attribute-frame postings indexes

Note that we're making sure the frames actually exist.

*/

insert into disk_index.frame_attribute_posting(frame_id, attribute_id)
    select frame_id, attribute.id
    from attribute_posting post
    inner join disk_index.attribute
        on attribute.type = post.type
        and attribute.value = post.value
    inner join disk_index.frame
        on frame.id = post.frame_id
    order by frame_id, attribute.id;

insert into disk_index.attribute_frame_posting(frame_id, attribute_id)
    select frame_id, attribute.id
    from attribute_posting post
    inner join disk_index.attribute
        on attribute.type = post.type
        and attribute.value = post.value
    inner join disk_index.frame
        on frame.id = post.frame_id
    order by attribute.id, frame_id;


/* Update the statistics by combining on-disk, deleted and new values */
with update_stat as (
    select
        v.id as term_id,
        fields.id as field_id,
        frequency,
        frames_occuring,
        documents_occuring
    from main.term_statistics stats
    inner join disk_index.vocabulary v
        on v.term = stats.term
    inner join disk_index.unstructured_field fields
        on fields.name = stats.field_name

    union all

    select *
    from deleted_term_statistics
),
updated_term as (
select distinct term_id, field_id from update_stat
)
insert or replace into disk_index.term_statistics
    select
        term_id,
        field_id,
        sum(frequency),
        sum(frames_occuring),
        sum(documents_occuring)
    from (select *
          from update_stat
          union all
          select *
          from disk_index.term_statistics
          where term_id in (select term_id from updated_term)
            and field_id in (select field_id from updated_term)
    )
    group by term_id, field_id;


-- Update settings
insert into disk_index.setting
    select *
    from setting;

-- Update the revision number of the database
insert into index_revision(added_document_count, deleted_document_count, added_frame_count)
    values(:added, :deleted, :added_frames);

-- Update the field statistics
insert or replace into field_statistics(field_id, frame_count)
    select field_id, count(*)
    from disk_index.frame
    group by field_id;

-- Update the plugins
create table delete_plugin_id as
    select registry.plugin_id
    from disk_index.plugin_registry registry
    inner join delete_plugin del
        on del.plugin_type = registry.plugin_type
        -- If settings is not supplied, delete all plugins of that type
        and (del.settings = registry.settings or del.settings is NULL)
;

delete from disk_index.plugin_data
where plugin_id in (select * from delete_plugin_id);

delete from disk_index.plugin_registry
where plugin_id in (select * from delete_plugin_id);

-- Insert header record for plugins
insert into disk_index.plugin_registry(plugin_type, settings)
    select *
    from plugin_registry;

-- Clear old data for updated plugins and insert new data
delete from disk_index.plugin_data
    where plugin_id in (
        select plugin_id
        from plugin_registry
        inner join disk_index.plugin_registry
            using(plugin_type, settings)
        );

insert into disk_index.plugin_data
    select plugin_id, plugin_data.key, plugin_data.value
    from plugin_data
    inner join disk_index.plugin_registry
        using(plugin_type, settings)
;

-- Return plugin ID's of new and inserted plugins
select plugin_type, settings, plugin_id
from plugin_registry
inner join disk_index.plugin_registry disk_reg
    using (settings, plugin_type);


delete from structured_field;
delete from unstructured_field;
delete from document;
delete from document_data;
delete from deleted_document;
delete from frame;
delete from setting;
delete from stage_posting;
delete from plugin_data;
delete from plugin_registry;
delete from delete_plugin;
drop table term_statistics;
drop index term_idx;

"""



# Copyright (c) 2012-2014 Kapiche Limited
# Author: Kris Rogers <kris@kapiche.com>, Ryan Stuart <ryan@kapiche.com>
"""
Indexes (see :mod:`caterpillar.processing.index`) in caterpillar must have a :class:`.Schema`. This module defines that
schema and also provides a bunch of utility functions for working with schemas and csv.

"""
from __future__ import absolute_import, division

import csv
import re
import sys

import regex

from caterpillar.processing.analysis.analyse import BiGramAnalyser, EverythingAnalyser, \
    DefaultAnalyser, DateTimeAnalyser
from caterpillar.processing.analysis.tokenize import Token


class FieldConfigurationError(Exception):
    pass


class FieldType(object):
    """
    Represents a field configuration. :class:`.Schema`s are built out of fields.

    The FieldType object controls how a field is analysed via the ``analyser``
    (:class:`caterpillar.processing.analysis.Analyser`) attribute.

    If you don't provide an analyser for your field, it will default to a
    :class:``caterpillar.processing.analysis.EverythingAnalyser``.

    """
    # Convenience hash of operators -> methods
    FIELD_OPS = {'<': 'lt', '<=': 'lte', '>': 'gt', '>=': 'gte', '*=': 'equals_wildcard', '=': 'equals'}

    def __init__(self, analyser=EverythingAnalyser(), indexed=False, categorical=False, stored=True):
        """
        Create a new field.

        ``analyser`` (:class:`caterpillar.processing.analysis.Analyser`) is the analyser used for this field.
        ``indexed`` (bool) indicates if this field should be indexed or not. ``categorical`` (bool) indicates if this
        field is categorical or not. Categorical fields only support indexing for the purpose of searching and do not
        collect full statistics such as positions and associations. ``stored`` (bool) says if this field should be
        stored or not.

        """
        self._analyser = analyser
        self._indexed = indexed
        self._categorical = categorical
        self._stored = stored

    def analyse(self, value):
        """Analyse ``value``, returning a :class:`caterpillar.processing.analysis.tokenize.Token` generator."""
        for token in self._analyser.analyse(value):
            yield token

    @property
    def categorical(self):
        return self._categorical

    @property
    def indexed(self):
        return self._indexed

    @property
    def stored(self):
        return self._stored

    def evaluate_op(self, operator, value1, value2):
        """
        Evaluate ``operator`` (str from :const:`FieldType.FIELD_OPS`) on operands ``value1`` and ``value2``.

        """
        op_method = getattr(self, FieldType.FIELD_OPS[operator])
        return op_method(value1, value2)

    def equals(self, value1, value2):
        """Returns whether ``value1`` is equal to ``value2``."""
        raise NotImplementedError('Equality operator not supported for field type {}.'.format(self.__class__.__name__))

    def equals_wildcard(self, value, wildcard_value):
        """Returns whether ``value`` matches regex ``wildcard_value``."""
        raise NotImplementedError('Wildcard equality operator not supported for field type {}.'
                                  .format(self.__class__.__name__))

    def gt(self, value1, value2):
        """Returns whether ``value1`` is greater than ``value2``."""
        raise NotImplementedError('Greater-than operator not supported for field type {}.'
                                  .format(self.__class__.__name__))

    def gte(self, value1, value2):
        """Returns whether ``value1`` is greater than or equal to ``value2``."""
        raise NotImplementedError('Greater-than-or-equals operator not supported for field type {}.'
                                  .format(self.__class__.__name__))

    def lt(self, value1, value2):
        """Returns whether ``value1`` is less than ``value2``."""
        raise NotImplementedError('Less-than operator not supported for field type {}.'
                                  .format(self.__class__.__name__))

    def lte(self, value1, value2):
        """Returns whether ``value1`` is less than or equal to ``value2``."""
        raise NotImplementedError('Less-than-or-equals operator not supported for field type {}.'
                                  .format(self.__class__.__name__))


class CategoricalFieldType(FieldType):
    """Represents a categorical field type. Categorical fields can extend this class for convenience."""
    def __init__(self, analyser=EverythingAnalyser(), indexed=False, stored=True):
        super(CategoricalFieldType, self).__init__(analyser=analyser, indexed=indexed, categorical=True, stored=stored)

    def value_of(self, raw_value):
        """Return the value of ``raw_value`` after being processed by this field type's analyse method."""
        return list(self.analyse(raw_value))[0].value

    def equals(self, value1, value2):
        return self.value_of(value1) == self.value_of(value2)


class ID(CategoricalFieldType):
    """
    Configured field type that indexes the entire value of the field as one token. This is useful for data you don't
    want to tokenize, such as the path of a file.

    """
    def __init__(self, indexed=False, stored=True):
        super(ID, self).__init__(indexed=indexed, stored=stored)


class NUMERIC(CategoricalFieldType):
    """Special field type that lets you index ints or floats."""
    TYPES = (int, float)

    def __init__(self, indexed=False, stored=True, num_type=int, default_value=None):
        """Create new NUMERIC instance with type ``num_type`` (float or int) and default_value (float or int)."""
        if num_type not in NUMERIC.TYPES:
            raise ValueError("Invalid num_type '{}'".format(num_type))
        self._num_type = num_type
        self._default_value = default_value
        super(NUMERIC, self).__init__(analyser=None, indexed=indexed, stored=stored)

    def analyse(self, value):
        try:
            yield Token(self._num_type(value))
        except (TypeError, ValueError) as e:
            if value is None or len(value) == 0:
                yield Token(self._default_value)
            else:
                raise e

    def gt(self, value1, value2):
        return self.value_of(value1) > self.value_of(value2)

    def gte(self, value1, value2):
        return self.value_of(value1) >= self.value_of(value2)

    def lt(self, value1, value2):
        return self.value_of(value1) < self.value_of(value2)

    def lte(self, value1, value2):
        return self.value_of(value1) <= self.value_of(value2)


class BOOLEAN(CategoricalFieldType):
    """
    bool field type that lets you index boolean values (True and False).

    The field converts the bool values to text for you before indexing.

    """
    def __init__(self, indexed=False, stored=True):
        super(BOOLEAN, self).__init__(analyser=None, indexed=indexed, stored=stored)

    def analyse(self, value):
        yield Token(bool(value))


class TEXT(FieldType):
    """Configured field type for text fields."""
    def __init__(self, analyser=DefaultAnalyser(), indexed=True, stored=True):
        """
        Create a text field with ``analyser`` (:class:`caterpillar.processing.analysis.Analyser`) default to
        :class:`caterpillar.processing.analysis.DefaultAnalyzer`.

        """
        super(TEXT, self).__init__(analyser=analyser, indexed=indexed, categorical=False, stored=stored)


class CATEGORICAL_TEXT(CategoricalFieldType):
    """Configured field type for categorical text fields."""
    def __init__(self, indexed=False, stored=True):
        super(CATEGORICAL_TEXT, self).__init__(indexed=indexed, stored=stored)

    def analyse(self, value):
        yield Token(value.strip())

    def equals_wildcard(self, value, wildcard_value):
        return re.match(wildcard_value, value) is not None


class DATETIME(FieldType):
    """Field type for datetimes.

    Datetimes are stored as ISO8601 strings down to a resolution of 1 second. All datetimes are either datetime
    unaware (no UTC offset) or datetime aware and stored as UTC (explicit offset +00:00). This means all datetimes
    are lexicographically comparable.

    """

    def __init__(self, analyser=DateTimeAnalyser(), indexed=False, stored=True):
        super(DATETIME, self).__init__(analyser=analyser, indexed=indexed, stored=stored, categorical=True)

    def value_of(self, raw_value):
        """Return the value of ``raw_value`` after being processed by this field type's analyse method."""
        return list(self.analyse(raw_value))[0].value

    def gt(self, value1, value2):
        return self.value_of(value1) > self.value_of(value2)

    def gte(self, value1, value2):
        return self.value_of(value1) >= self.value_of(value2)

    def lt(self, value1, value2):
        return self.value_of(value1) < self.value_of(value2)

    def lte(self, value1, value2):
        return self.value_of(value1) <= self.value_of(value2)

    def equals(self, value1, value2):
        return self.value_of(value1) == self.value_of(value2)


class Schema(object):
    """
    Represents the collection of fields in an index. Maps field names to FieldType objects which define the behavior of
    each field.

    Low-level parts of the index use field numbers instead of field names for compactness. This class has several
    methods for converting between the field name, field number, and field object itself.

    """
    def __init__(self, **fields):
        """
        All keyword arguments to the constructor are treated as ``fieldname = fieldtype`` pairs. The fieldtype can be an
        instantiated FieldType object, or a FieldType sub-class (in which case the Schema will instantiate it with the
        default constructor before adding it).

        For example::

            s = Schema(content = TEXT, title = TEXT(stored = True), tags = KEYWORD(stored = True))

        """

        self._fields = {}

        for name in sorted(fields.keys()):
            self.add(name, fields[name])

    def __iter__(self):
        """Returns the field objects in this schema."""
        return iter(self._fields.values())

    def __getitem__(self, name):
        """Returns the field associated with the given field name."""
        if name in self._fields:
            return self._fields[name]

        raise KeyError("No field named {}".format(name))

    def __len__(self):
        """Returns the number of fields in this schema."""
        return len(self._fields)

    def __contains__(self, field_name):
        """Returns True if a field by the given name is in this schema."""
        # Defined in terms of __getitem__ so that there's only one method to override to provide dynamic fields
        try:
            field = self[field_name]
            return field is not None
        except KeyError:
            return False

    def items(self):
        """Returns a list of ``("field_name", field_object)`` pairs for the fields in this schema."""
        return sorted(self._fields.items())

    def names(self):
        """Returns a list of the names of the fields in this schema."""
        return sorted(self._fields.keys())

    def get_indexed_text_fields(self):
        """Returns a list of the indexed text fields."""
        return [name for name, field in self._fields.iteritems() if field.indexed and type(field) == TEXT]

    def get_indexed_structured_fields(self):
        """Returns a list of the indexed structured (non-text) fields."""
        return [name for name, field in self._fields.iteritems() if field.indexed and type(field) != TEXT]

    def add(self, name, field_type):
        """
        Adds a field to this schema.

        ``name`` (str) is the name of the field. ``fieldtype`` (:class:`FieldType`) is either
        instantiated FieldType object, or a FieldType subclass. If you pass an instantiated object, the schema will use
        that as the field configuration for this field. If you pass a FieldType subclass, the schema will automatically
        instantiate it with the default constructor.

        """
        # Check field name
        if name.startswith("_"):
            raise FieldConfigurationError("Field names cannot start with an underscore")
        if name in self._fields:
            raise FieldConfigurationError("Schema already has a field {}".format(name))

        # If the user passed a type rather than an instantiated field object,
        # instantiate it automatically
        if type(field_type) is type:
            try:
                field_type = field_type()
            except:
                e = sys.exc_info()[1]
                raise FieldConfigurationError("Error: {} instantiating field {}: {}".format(e, name, field_type))

        if not isinstance(field_type, FieldType):
            raise FieldConfigurationError("{} is not a FieldType object".format(field_type))

        self._fields[name] = field_type


class ColumnDataType(object):
    """
    A ColumnDataType object identifies different data types in a CSV column.

    There are five possible data types:
    FLOAT -- A floating point number. Should be a string in a format that float() can parse.
    INTEGER -- A integer. Should be in a format that int() can parse.
    STRING -- A string type. This type ISN'T analysed. It is just stored.
    TEXT -- A string type. This is like STRING but it IS analysed and stored (it is used to generate a frame stream).
    IGNORE -- Ignore this column.

    """
    FLOAT = 1
    INTEGER = 2
    STRING = 3
    TEXT = 4
    IGNORE = 5


class ColumnSpec(object):
    """
    A ColumnSpec object represents information about a column in a CSV file.

    This includes the column's name and its type.

    """

    def __init__(self, name, type):
        """
        Create a new ColumnSpec. A ColumnSpec must have a name and a type.

        Required Arguments:
        name -- A string name for this column.
        type -- The type of this column as a ColumnDataType object.

        """
        self.name = name
        self.field_name = name.replace(' ', '')
        self.type = type


class CsvSchema(object):
    """
    This class represents the schema required to process a particular CSV data file.

    Required Arguments:
    columns -- A list of ``ColumnSpec`` objects to define how the data should be processed.
    has_header -- A boolean indicating whether the first row of the file contains headers.
    dialect -- The dialect to use when parsing the file.

    Optional Arguments:
    sample_rows -- A list of row data that was used to generate the schema.

    """
    def __init__(self, columns, has_header, dialect, sample_rows=[]):
        self.columns = columns
        self.has_header = has_header
        self.dialect = dialect
        self.sample_rows = sample_rows

    def as_index_schema(self, bi_grams=None, stopword_list=None):
        """
        Return a representation of this ``CsvSchema`` in the form of a ``Schema`` instance that can be
        used for generating a text index.

        Optional Arguments:
        bi_grams -- A list of bi-grams to use for TEXT columns.
        stopword_list -- A list of stop words to use instead of default English list.

        """
        schema = Schema()
        for col in self.columns:

            if col.type == ColumnDataType.IGNORE:
                continue

            elif col.type == ColumnDataType.FLOAT:
                schema.add(col.field_name, NUMERIC(num_type=float))

            elif col.type == ColumnDataType.INTEGER:
                schema.add(col.field_name, NUMERIC(num_type=int))

            elif col.type == ColumnDataType.STRING:
                schema.add(col.field_name, CATEGORICAL_TEXT)

            elif col.type == ColumnDataType.TEXT:
                if bi_grams is not None:
                    schema.add(col.field_name, TEXT(analyser=BiGramAnalyser(bi_grams, stopword_list=stopword_list)))
                else:
                    schema.add(col.field_name, TEXT(analyser=DefaultAnalyser(stopword_list=stopword_list)))

        return schema

    def map_row(self, row):
        """
        Convert a row into dict form to make it easier to index.

        Required Arguments:
        row -- A list of values for a row.

        """
        row_data = {}
        i = 0
        for col in self.columns:
            if col.type != ColumnDataType.IGNORE:
                row_data[col.field_name] = row[i]
            i += 1
        return row_data


AVG_WORDS_TEXT = 5  # Minimum number of average words per row to consider a column as text
NUM_PEEK_ROWS_CSV = 20  # Number of csv rows to consider when generating automatic schema


def generate_csv_schema(csv_file, delimiter=',', encoding='utf8'):
    """
    Attempt to generate a schema for the csv file automatically.

    Required Arguments:
    csv_file -- The CSV file to generate a schema for.

    Optional Arguments:
    delimiter -- CSV delimiter character.
    encoding -- Character encoding of the file.

    Returns a 2-tuple containing the generated schema and the sample rows used to generate the schema.

    """
    dialect = csv.excel

    # Check for headers
    has_header = csv_has_header(csv_file, dialect)

    # Now actually read the file
    csv_file.seek(0)
    reader = csv.reader(csv_file, dialect)
    headers = []
    if has_header:
        headers = reader.next()

    # Collect column statistics
    sample_rows = []
    column_stats = {}
    for i in range(NUM_PEEK_ROWS_CSV):
        try:
            row = reader.next()
        except StopIteration:
            # No more rows
            break
        sample_rows.append(row)
        for j in range(len(row)):
            col = row[j]
            if j not in column_stats:
                column_stats[j] = {'total_words': 0}
            column_stats[j]['total_words'] += len(regex.findall(r'\w+', col))

    # Define columns and generate schema
    columns = []
    for index, stats in column_stats.items():
        name = None
        if headers and index < len(headers):
            name = unicode(headers[index], encoding, errors='ignore')
        if name is None or len(name) == 0:
            name = str(index + 1)
        if stats['total_words'] / NUM_PEEK_ROWS_CSV >= AVG_WORDS_TEXT:
            # Enough words for a text column
            columns.append(ColumnSpec(name, ColumnDataType.TEXT))
        else:
            # Didn't match anything, default to IGNORE
            columns.append(ColumnSpec(name, ColumnDataType.IGNORE))

    return CsvSchema(columns, has_header, dialect, sample_rows)


MAX_HEADER_SIZE_PERCENTAGE = 0.33  # Maximum size for header row as a percentage of the average row size


def csv_has_header(csv_file, dialect, num_check_rows=50):
    """
    Custom heuristic for recognising header in CSV files. Intended to be used as an alternative
    for the ``csv.Sniffer.has_header`` method which doesn't work well for mostly-text CSV files.

    The heuristic we use simply checks the total size of the header row compared to the average row size for
    the following sample rows. If a large discrepancy is found, we assume that the first row contains headers.

    Required Arguments:
    csv_file -- The csv data file to analyse.
    dialect -- CSV dialect to use.

    Optional Arguments:
    num_check_rows -- The number of rows to analyse (defaults to 50).

    """
    reader = csv.reader(csv_file, dialect)
    header = reader.next()  # assume first row is header
    header_size = sum([len(col) for col in header])

    # Compute average row size
    total_row_size = 0
    checked = 0
    for row in reader:
        if checked == num_check_rows:
            break
        total_row_size += sum([len(col) for col in row])
        checked += 1
    avg_row_size = total_row_size / checked

    return header_size / avg_row_size <= MAX_HEADER_SIZE_PERCENTAGE



# Copyright (c) 2012-2014 Kapiche Limited
# Author: Ryan Stuart <ryan@kapiche.com>
import os
from abc import abstractproperty

version_file = open(os.path.join(os.path.dirname(__file__), 'VERSION'), 'r')
full_version = version_file.read().strip().split('-')
version_file.close()

VERSION = full_version[0]  # Major version number, X.Y
RELEASE = full_version[1] if len(full_version) > 1 else ''  # Release name


def abstract_method_tester(abc):
    """Utility function to test abstract classes"""

    def tester(self, abc_method_name):
        abc_method = abc.__dict__[abc_method_name]
        if isinstance(abc_method, abstractproperty):
            abstractproperty(self)
        else:
            args = (None,) * (abc_method.func_code.co_argcount - 1)
            abc_method(self, *args)

    methods = {}
    methods.update((
        n, (lambda n: lambda self: tester(self, n))(n))
        for n in ['__init__'] + list(abc.__abstractmethods__)
    )
    subinstance = type(abc.__name__ + "_test", (abc,), methods)()
    for method in methods.itervalues():
        method(subinstance)



# Copyright (c) 2012-2014 Kapiche Limited
# Author: Kris Rogers <kris@kapiche.com>, Ryan Stuart <ryan@kapiche.com>
import os
import re

MIN_WORD_SIZE = 3  # length of the smallest possible word
APOSTROPHES = u'\u0027\u2019\u02BC\u02BB\u055A\uA78B\uA78C\uFF07'
APOS_RE = re.compile(u'[%s]' % APOSTROPHES)


def parse_stopwords(stopwords_file):
    """
    Parse stopwords from a plain text file.

    Expects a single stopword on every line.

    If a stopword has some flavour of an apostrophe character, we make sure we insert a version of that stopword with
    every type of unicode apostrophe character.

    """
    stopwords = []
    for line in stopwords_file:
        stopword = line.strip()
        if APOS_RE.search(stopword) is not None:
            for apos in APOSTROPHES:
                stopwords.append(APOS_RE.sub(apos, stopword))
        else:
            stopwords.append(stopword)

    return stopwords


# Stopword lists
ENGLISH = None
with open(os.path.join(os.path.dirname(__file__), '../../', 'resources', 'stopwords-english.txt')) as stopwords_file:
    ENGLISH = parse_stopwords(stopwords_file)
ENGLISH_TEST = None
with open(os.path.join(os.path.dirname(__file__), '../../', 'resources',
                       'stopwords-english-test.txt')) as stopwords_file:
    ENGLISH_TEST = parse_stopwords(stopwords_file)



# Copyright (c) Kapiche Limited
# Author: Sam Hames <sam.hames@kapiche.com>

"""
Operators for creating complex searches by composition of simple searches.

The :class:`IndexReader` exposes :meth:`.filter` and :meth:`.filter_and_rank` which provide fast low
level access to frames and documents in an index by both structured and unstructured data. Not all
search needs are met by this method, however, so the methods in this module allow composition and
ranking of resultsets.

A resultset is a dictionary mapping a frame_id or document_id to a list of scores:
    {1: [10], 211: [4], 102: [8]}

Results from `IndexReader.filter` will have the TF-IDF score of matching terms (or 0 if only
metadata was used to filter) - these primitive queries have only a single score. When combining
resultsets a list of scores for each matching result is accumulated.

This module provides three operators for combining resultsets: match_all for taking the intersection
of multiple resultsets, match_any for taking the union and exclude for removing results from one set
that occur in a different set.

Examples
    >>> with IndexReader('/path/to/index') as r:
    ...    result1 = r.filter(must=[('cat', 'cats', 'dog', 'dogs'), 'pet'], return_documents=False)
    ...    result2 = r.filter(should=['petshop', 'kennel'], return_documents=False)
    ...    result3 = r.filter(metadata={'timestamp': {'>=': '2016-01-01T00:00:00'}}, return_documents=False)
    ...
    ...    # Intersection of all result sets
    ...    result4 = match_all(result1, result2, result3)
    ...
    ...    # Union of result sets
    ...    result5 = match_any(result1, result2)
    ...
    ...    # Remove anything matching result3 from result5
    ...    result6 = exclude(result5, result3)
    ...
    ...    # Score and rank result5, taking the maximum of each score from the low level matches and
    ...    # returning only the first 10 keys
    ...    scored = score_and_rank(result5, aggregator=max, start=0, limit=10):
    ...
    ...    # Score and rank result6, taking the sum of all the individual scores from each match.
    ...    scored2 = score_and_rank(result6, aggregator=sum)
    ...
    ...    # Load the frames corresponding to scored:
    ...    frames = list(r.get_frames(None, frame_ids=[i[0] for i in scored2]))
    ...

"""


def match_all(*result_sets):
    """
    Take the intersection of all the result_sets.

    If a key is missing from any of the input sets, it will not be present in the output.

    """
    # View keys is like a set object of the dictionary keys - this will become keys() in Python 3
    keep_keys = result_sets[0].viewkeys()

    for d in result_sets[1:]:
        keep_keys &= d.viewkeys()

    output = {key: [i for d in result_sets for i in d[key]] for key in keep_keys}

    return output


def match_any(*result_sets):
    """
    Take the union of all the result_sets.

    If a key is present in any of the input sets, it will be present in the output. For any key in
    the output set that is missing from a result set, a score of 0 will be assigned.

    """

    # View keys is like a set object of the dictionary keys - this will become keys() in Python 3
    all_keys = result_sets[0].viewkeys()

    for d in result_sets[1:]:
        all_keys |= d.viewkeys()

    # Note that explicit 0 scores are assigned if a key is missing, so that each score set has one
    # entry per query component.
    output = {key: [i for d in result_sets for i in d.get(key, [0])] for key in all_keys}

    return output


def exclude(result_set, *exclusions):
    """Match keys in result_set, but only if they do not occur in any of the exclusions."""
    keep_keys = result_set.viewkeys()

    for exclusion in exclusions:
        keep_keys -= exclusion.viewkeys()

    return {key: result_set[key] for key in keep_keys}


def boost(result_set, boost):
    """Multiply each individual score for this result set by boost."""
    return {key: [i * boost for i in value] for key, value in result_set.items()}


def score_and_rank(result_set, aggregator=sum, start=0, limit=0):
    """
    Aggregate the scores for each match, then return tuples of (id, score) in descending order.

    The aggregator function should take a sequence of numbers and return a single number.

    """
    aggregated = sorted(
        ((object_id, aggregator(value)) for object_id, value in result_set.items()),
        key=lambda x: (x[1], -x[0]),  # Deterministic sort on keys if scores are identical
        reverse=True
    )

    if limit > 0:
        return aggregated[start:start + limit]
    else:
        return aggregated



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from caterpillar.processing.analysis import stopwords
from caterpillar.processing.analysis.analyse import Analyser
from caterpillar.processing.analysis.filter import OuterPunctuationFilter, StopFilter, PositionalLowercaseWordFilter, \
    BiGramFilter
from caterpillar.processing.analysis.filter import PossessiveContractionFilter
from caterpillar.processing.analysis.tokenize import SimpleWordTokenizer


class TestAnalyser(Analyser):
    _tokenizer = SimpleWordTokenizer(detect_compound_names=True)

    def __init__(self, stopword_list=None):
        super(TestAnalyser, self).__init__()
        if stopword_list is None:
            stopword_list = stopwords.ENGLISH_TEST

        self._filters = [
            OuterPunctuationFilter(leading_allow=['@', '#']),
            PossessiveContractionFilter(),
            StopFilter(stopword_list, minsize=stopwords.MIN_WORD_SIZE),
            PositionalLowercaseWordFilter(0),
        ]

    def get_tokenizer(self):
        return self._tokenizer

    def get_filters(self):
        return self._filters


class TestBiGramAnalyser(Analyser):
    _tokenizer = SimpleWordTokenizer(detect_compound_names=True)

    def __init__(self, bi_grams, stopword_list=None):
        if stopword_list is None:
            stopword_list = stopwords.ENGLISH_TEST
        self._filters = [
            OuterPunctuationFilter(leading_allow=['@', '#']),
            PossessiveContractionFilter(),
            StopFilter(stopword_list, minsize=stopwords.MIN_WORD_SIZE),
            PositionalLowercaseWordFilter(0),
            BiGramFilter(bi_grams),
        ]

    def get_tokenizer(self):
        return self._tokenizer

    def get_filters(self):
        return self._filters



# Copyright (c) 2012-2014 Kapiche Limited
# Author: Ryan Stuart <ryan@kapiche.com>
"""Tools for filtering tokens."""
import regex


class Filter(object):
    """
    Base class for Filter objects.

    A Filter subclass must implement a filter() method that takes a single argument, which is an iterator of Token
    objects, and yield a series of Token objects in return.

    """
    def filter(self, tokens):
        """
        Return filtered tokens from the tokens iterator.

        """
        raise NotImplementedError


class PassFilter(Filter):
    """
    An identity filter: passes the tokens through untouched.

    """
    def filter(self, tokens):
        return tokens


class LowercaseFilter(Filter):
    """
    Uses unicode.lower() to lowercase token text.

    >>> rext = RegexTokenizer()
    >>> stream = rext("This is a TEST")
    >>> [token.value for token in LowercaseFilter(stream)]
    ["this", "is", "a", "test"]

    """
    def filter(self, tokens):
        for t in tokens:
            t.value = t.value.lower()
            yield t


class PositionalLowercaseWordFilter(Filter):
    """
    Uses unicode.lower() to lowercase single word tokens that appear to be titles and are in a certain position.

    Only applies the filter if the token only contains only 1 word (ie no spaces) and unicode.istitle() returns true.
    Used internally to force the case of words appearing at the start of a sentence to lower if it isn't part of a
    compound name.

    >>> rext = RegexTokenizer()
    >>> stream = rext("This is a TEST")
    >>> [token.value for token in LowercaseFilter(0).filter()]
    ["this", "is", "a", "TEST"]

    Required Arguments:
    position -- A 0 based int index the token needs to be in to apply this filter.

    """
    def __init__(self, position):
        self._position = position

    def filter(self, tokens):
        for t in tokens:
            if t.position == self._position and t.value.istitle() and ' ' not in t.value:
                t.value = t.value.lower()
            yield t


class StopFilter(Filter):
    """
    Marks "stop" words (words too common to index) in the stream.

    >>> rext = RegexTokenizer()
    >>> stream = rext("this is a test")
    >>> stopper = StopFilter()
    >>> [token.value for token in stopper(stream)]
    ["this", "test"]

    Required Arguments
    stoplist -- A list of stop words lower cased.

    Optional Arguments
    minsize -- An int indicating the smallest acceptable word length.

    """
    def __init__(self, stoplist, minsize=3):
        self._stoplist = {s.lower(): None for s in stoplist}
        self._minsize = minsize

    def filter(self, tokens):
        for t in tokens:
            if len(t.value) < self._minsize or t.value.lower() in self._stoplist:
                t.stopped = True
            yield t


class BiGramFilter(Filter):
    """
    Identifies bi-grams in a token stream from a given list.

    >>> rext = RegexTokenizer()
    >>> stream = rext("this is a bigram")
    >>> ngrams = BiGramFilter(["a bigram"])
    >>> [token.value for token in bigrams(stream)]
    ["this", "is, "a bigram"]

    Required Arguments
    bi_grams -- A list of n-gram strings to match.

    """
    def __init__(self, bi_grams):
        self._bi_grams = bi_grams

    def filter(self, tokens):
        prev_token = None
        for t in tokens:
            if t.value[0].isupper() or t.stopped:
                # Names and stopwords can't be bi-grams
                if prev_token:
                    yield prev_token
                    prev_token = None
                yield t
                continue
            elif prev_token:
                bi_gram = u"{} {}".format(prev_token.value, t.value)
                if bi_gram in self._bi_grams:
                    yield t.update(bi_gram, position=prev_token.position, index=(prev_token.index[0], t.index[1]))
                    prev_token = None
                    continue
                else:
                    yield prev_token
            prev_token = t.copy()
        if prev_token:  # Don't forget the last token!!
            yield prev_token


class PotentialBiGramFilter(Filter):
    """
    Identifies bi-grams in a token stream along with regular tokens.

    Potential bi-grams won't include stopped tokens or names.

    WARNING: This filter differs from most other filters in that it returns a list of token objects, not just single
    tokens. This is done purely for performance. If it finds a non-bi-gram, a list of just a single token is returned.

    """
    def filter(self, tokens):
        prev_token = None
        for t in tokens:
            if t.value[0].isupper() or t.stopped:
                # Skip names and stopwords
                if prev_token:
                    yield [prev_token]
                    prev_token = None
                yield [t]
                continue
            elif prev_token:
                yield [prev_token, t]
            prev_token = t.copy()
        if prev_token:
            yield [prev_token]


class SubstitutionFilter(Filter):
    """
    Performs a regular expression substitution on the token text.

    This is especially useful for removing text from tokens, for example hyphens::

        ana = RegexTokenizer(r"\\S+") | SubstitutionFilter("-", "")

    Because it has the full power of the ``regex.sub()`` method behind it, this filter can perform some fairly complex
    transformations. For example, to take tokens like ``'a=b', 'c=d', 'e=f'`` and change them to ``'b=a', 'd=c',
    'f=e'``::

    >>> sf = SubstitutionFilter("([^/]*)/(./*)", r"\\2/\\1")

    Required Arguments
    pattern -- A pattern string or compiled regular expression object describing the text to replace.
    replacement -- A string of substitution text.

    """
    def __init__(self, pattern, replacement):
        self.pattern = regex.compile(pattern, flags=regex.UNICODE | regex.DOTALL | regex.VERSION1)
        self.replacement = replacement

    def filter(self, tokens):
        pattern = self.pattern
        replacement = self.replacement

        for t in tokens:
            t.value = pattern.sub(replacement, t.value)
            if t:
                yield t


class SearchFilter(Filter):
    """
    Performs a regular expression search on the token text and uses match group 0 as the token value. If no match is
    found then the token is skipped.

    :param str pattern: The pattern used in the re search.
    """
    def __init__(self, pattern):
        self.pattern = regex.compile(pattern, flags=regex.UNICODE | regex.DOTALL | regex.VERSION1)

    def filter(self, tokens):
        pattern = self.pattern

        for t in tokens:
            match = pattern.search(t.value)
            if match:
                t.value = match.group(0)
                yield t


class OuterPunctuationFilter(SearchFilter):
    """
    Remove some or all leading and trailing punctuation from tokens.

    User of this filter can specify which leading and/or trailing punctuation to leave in tact.
    """
    def __init__(self, leading_allow=None, trailing_allow=None):
        """
        :param list leading_allow: The leading punctuation characters to allow.
        :param list trailing_allow: The trailing punctuation characters to allow.
        """
        leading_pattern = '' if not leading_allow else r'[%s]*' % regex.escape("".join(leading_allow))
        trailing_pattern = '' if not trailing_allow else r'[%s]' % regex.escape("".join(trailing_allow))
        if trailing_pattern:
            super(OuterPunctuationFilter, self).__init__('%s[^\W_]+(?:$|.*[^\W_]%s*|%s*)' %
                                                         (leading_pattern, trailing_pattern, trailing_pattern))
        else:
            super(OuterPunctuationFilter, self).__init__('%s[^\W_](?:$|.*[^\W_])' % leading_pattern)


class PossessiveContractionFilter(SubstitutionFilter):
    """
    Removes possessive contractions from tokens.

    Is fairly robust is that is uses all know unicode apostrophe characters except U+02EE. See
    `http://en.wikipedia.org/wiki/Apostrophe#Unicode`_.
    """
    def __init__(self):
        super(PossessiveContractionFilter, self).__init__(u"[\u0027\u2019\u02BC\u02BB\u055A\uA78B\uA78C\uFF07]s$", "")



# Copyright (c) 2012-2014 Kapiche Limited
# Author: Ryan Stuart <ryan@kapiche.com>
"""Tools to perform analysis of text streams (aka tokenizing and filtering)."""
from caterpillar.processing.analysis import stopwords
from caterpillar.processing.analysis.filter import StopFilter, PositionalLowercaseWordFilter, BiGramFilter, \
    PotentialBiGramFilter, OuterPunctuationFilter, PossessiveContractionFilter
from caterpillar.processing.analysis.tokenize import EverythingTokenizer, \
    SimpleWordTokenizer, DateTimeTokenizer


class Analyser(object):
    """
    Abstract base class for an analyser.

    All analysers are a combination of a tokenizer and 0 or more filters. This class accesses the tokenizer by calling
    self.get_tokenizer() and the filters via self.get_filters(). You need to implement the get_tokenizer() method at
    a minimum.

    This class also defines the analyse() method which will call the tokenizer followed by the filters in order before
    finally returning the token.

    """
    def analyse(self, value):
        # Tokenize first
        token = self.get_tokenizer().tokenize(value)
        # Then filter
        if self.get_filters():
            for f in self.get_filters():
                token = f.filter(token)
        return token

    def get_tokenizer(self):
        raise NotImplementedError

    def get_filters(self):
        return None


class DefaultAnalyser(Analyser):
    """
    The default caterpillar ``Analyser`` which mostly splits on whitespace and punctuation, except for a few special
    cases, and removes stopwords.

    This analyzer uses a ``WordTokenizer`` in combination with a ``StopFilter`` and a ``PositionalLowercaseWordFilter``.

    Optional Arguments:
    stopword_list -- A list of stop words to override the default ``stopwords.ENGLISH`` one.

    """
    _tokenizer = SimpleWordTokenizer(detect_compound_names=True)

    def __init__(self, stopword_list=[], min_word_size=1):
        super(DefaultAnalyser, self).__init__()
        if stopword_list is None:
            stopword_list = stopwords.ENGLISH

        self._filters = [
            OuterPunctuationFilter(leading_allow=['@', '#']),
            PossessiveContractionFilter(),
            StopFilter(stopword_list, minsize=min_word_size),
            PositionalLowercaseWordFilter(0),
        ]

    def get_tokenizer(self):
        return self._tokenizer

    def get_filters(self):
        return self._filters


class BiGramAnalyser(Analyser):
    """
    A bi-gram ``Analyser`` that behaves exactly like the ``DefaultAnalyser`` except it also makes use of a
    ``BiGramFilter``.

    This analyser uses a ``WordTokenizer`` in combination with a ``StopFilter``, ``PositionalLowercaseWordFilter`` and a
    ``BiGramFilter``.

    Required Arguments
    bi_grams -- a list of string n-grams to match. Passed directly to ``BiGramFilter``.

    Optional Arguments:
    stopword_list -- A list of stop words to override the default English one.

    """
    _tokenizer = SimpleWordTokenizer(detect_compound_names=True)

    def __init__(self, bi_grams, stopword_list=None):
        if stopword_list is None:
            stopword_list = stopwords.ENGLISH
        self._filters = [
            OuterPunctuationFilter(leading_allow=['@', '#']),
            PossessiveContractionFilter(),
            StopFilter(stopword_list, minsize=stopwords.MIN_WORD_SIZE),
            PositionalLowercaseWordFilter(0),
            BiGramFilter(bi_grams),
        ]

    def get_tokenizer(self):
        return self._tokenizer

    def get_filters(self):
        return self._filters


class PotentialBiGramAnalyser(Analyser):
    """
    A PotentialBiGramAnalyser returns a list of possible bi-grams from a stream.

    This analyser uses a ``WordTokenizer`` in combination with a ``StopFilter``, ``PositionalLowercaseWordFilter`` and a
    ``PotentialBiGramFilter`` to generate a stream of possible bi-grams.

    """
    _tokenizer = SimpleWordTokenizer(detect_compound_names=True)
    _filters = [
        OuterPunctuationFilter(leading_allow=['@', '#']),
        PossessiveContractionFilter(),
        StopFilter(stopwords.ENGLISH, minsize=stopwords.MIN_WORD_SIZE),
        PositionalLowercaseWordFilter(0),
        PotentialBiGramFilter(),
    ]

    def get_tokenizer(self):
        return self._tokenizer

    def get_filters(self):
        return self._filters


class EverythingAnalyser(Analyser):
    """
    A EverythingAnalyser just returns the entire input string as a token.

    """
    _tokenizer = EverythingTokenizer()
    _filters = []

    def get_tokenizer(self):
        return self._tokenizer

    def get_filters(self):
        return self._filters


class DateTimeAnalyser(Analyser):
    """
    Analyser for ``DATETIME`` fields. Returns an ISO8601 format datetime string.

    """
    _filters = []

    def __init__(self, datetime_formats=None, ignore_tz=False):
        self.datetime_formats = datetime_formats
        self.ignore_tz = ignore_tz
        self._tokenizer = DateTimeTokenizer(datetime_formats, ignore_tz)

    def get_tokenizer(self):
        return self._tokenizer

    def get_filters(self):
        return self._filters



# Copyright (C) 2012-2014 Kapiche Limited
# Author: Ryan Stuart <ryan@kapiche.com>
import os
import shutil
import tempfile

import pytest


@pytest.fixture
def index_dir(request):
    path = tempfile.mkdtemp()

    def clean():
        shutil.rmtree(path)

    request.addfinalizer(clean)
    return os.path.join(path, "test_index")
